from __future__ import absolute_import, division, print_function

__metaclass__ = type
import six
from datetime import datetime
from decimal import Decimal
from random import randint

import pytest

from ..message import GenericContent
from ..serialization import AMQPReader, AMQPWriter, FrameSyntaxError


def assert_equal_binary(b, s):
    assert b == six.b(s)


class TestSerialization:
    def test_empty_writer(self):
        w = AMQPWriter()
        assert w.getvalue() == bytes()

    def test_single_bit(self):
        for val, check in [(True, '\x01'), (False, '\x00')]:
            w = AMQPWriter()
            w.write_bit(val)
            s = w.getvalue()

            assert_equal_binary(s, check)

            r = AMQPReader(s)
            assert r.read_bit() == val

    def test_multiple_bits(self):
        w = AMQPWriter()
        w.write_bit(True)
        w.write_bit(True)
        w.write_bit(False)
        w.write_bit(True)
        s = w.getvalue()

        assert_equal_binary(s, '\x0b')

        r = AMQPReader(s)
        assert r.read_bit() is True
        assert r.read_bit() is True
        assert r.read_bit() is False
        assert r.read_bit() is True

    def test_multiple_bits2(self):
        """
        Check bits mixed with non-bits
        """
        w = AMQPWriter()
        w.write_bit(True)
        w.write_bit(True)
        w.write_bit(False)
        w.write_octet(10)
        w.write_bit(True)
        s = w.getvalue()

        assert_equal_binary(s, '\x03\x0a\x01')

        r = AMQPReader(s)
        assert r.read_bit() is True
        assert r.read_bit() is True
        assert r.read_bit() is False
        assert r.read_octet() == 10
        assert r.read_bit() is True

    def test_multiple_bits3(self):
        """
        Check bit groups that span multiple bytes
        """
        w = AMQPWriter()

        # Spit out 20 bits
        for i in range(10):
            w.write_bit(True)
            w.write_bit(False)

        s = w.getvalue()

        assert_equal_binary(s, '\x55\x55\x05')

        r = AMQPReader(s)
        for i in range(10):
            assert r.read_bit() is True
            assert r.read_bit() is False

    # -------------
    # Octets

    def test_octet(self):
        for i in range(256):
            w = AMQPWriter()
            w.write_octet(i)
            b = w.getvalue()
            assert_equal_binary(b, chr(i))

            r = AMQPReader(b)
            assert r.read_octet() == i

    def test_octet_invalid(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_octet(-1)

    def test_octet_invalid2(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_octet(256)

    # -------------
    # Shorts

    def test_short(self):
        for i in range(256):
            val = randint(0, 65535)
            w = AMQPWriter()
            w.write_short(val)
            s = w.getvalue()

            r = AMQPReader(s)
            assert r.read_short() == val

    def test_short_invalid(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_short(-1)

    def test_short_invalid2(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_short(65536)

    # -------------
    # Longs

    def test_long(self):
        for i in range(256):
            val = randint(0, 4294967295)
            w = AMQPWriter()
            w.write_long(val)
            s = w.getvalue()

            r = AMQPReader(s)
            assert r.read_long() == val

    def test_long_invalid(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_long(-1)

    def test_long_invalid2(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_long(4294967296)

    # -------------
    # LongLongs

    def test_longlong(self):
        for i in range(256):
            val = randint(0, (2 ** 64) - 1)
            w = AMQPWriter()
            w.write_longlong(val)
            s = w.getvalue()

            r = AMQPReader(s)
            assert r.read_longlong() == val

    def test_longlong_invalid(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_longlong(-1)

    def test_longlong_invalid2(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_longlong(2 ** 64)

    # -------------
    # Shortstr

    def test_empty_shortstr(self):
        w = AMQPWriter()
        w.write_shortstr('')
        s = w.getvalue()

        assert_equal_binary(s, '\x00')

        r = AMQPReader(s)
        assert r.read_shortstr() == ''

    def test_shortstr(self):
        w = AMQPWriter()
        w.write_shortstr('hello')
        s = w.getvalue()
        assert_equal_binary(s, '\x05hello')

        r = AMQPReader(s)
        assert r.read_shortstr() == 'hello'

    def test_shortstr_unicode(self):
        w = AMQPWriter()
        w.write_shortstr('hello')
        s = w.getvalue()
        assert_equal_binary(s, '\x05hello')

        r = AMQPReader(s)
        assert r.read_shortstr() == 'hello'

    def test_long_shortstr(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_shortstr('x' * 256)

    def test_long_shortstr_unicode(self):
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_shortstr('\u0100' * 128)

    # -------------
    # Longstr

    def test_empty_longstr(self):
        w = AMQPWriter()
        w.write_longstr('')
        s = w.getvalue()

        assert_equal_binary(s, '\x00\x00\x00\x00')

        r = AMQPReader(s)
        assert r.read_longstr() == ''

    def test_longstr(self):
        val = 'a' * 512
        w = AMQPWriter()
        w.write_longstr(val)
        s = w.getvalue()

        assert_equal_binary(s, '\x00\x00\x02\x00' + ('a' * 512))

        r = AMQPReader(s)
        assert r.read_longstr() == str(val)

    def test_longstr_unicode(self):
        val = 'a' * 512
        w = AMQPWriter()
        w.write_longstr(val)
        s = w.getvalue()

        assert_equal_binary(s, '\x00\x00\x02\x00' + ('a' * 512))

        r = AMQPReader(s)
        assert r.read_longstr() == val

    # -------------
    # Table

    def test_table_empty(self):
        val = {}
        w = AMQPWriter()
        w.write_table(val)
        s = w.getvalue()

        assert_equal_binary(s, '\x00\x00\x00\x00')

        r = AMQPReader(s)
        assert r.read_table() == val

    def test_table(self):
        val = {'foo': 7}
        w = AMQPWriter()
        w.write_table(val)
        s = w.getvalue()

        assert_equal_binary(s, '\x00\x00\x00\x09\x03fooI\x00\x00\x00\x07')

        r = AMQPReader(s)
        assert r.read_table() == val

    def test_table_invalid(self):
        """
        Check that an un-serializable table entry raises a ValueError

        """
        val = {'test': object()}
        w = AMQPWriter()
        with pytest.raises(FrameSyntaxError):
            w.write_table(val)

    def test_table_multi(self):
        val = {
            'foo': 7,
            'bar': Decimal('123345.1234'),
            'baz': 'this is some random string I typed',
            'ubaz': 'And something in unicode',
            'dday_aniv': datetime(1994, 6, 6),
            'nothing': None,
            'more': {
                'abc': -123,
                'def': 'hello world',
                'now': datetime(2007, 11, 11, 21, 14, 31),
                'qty': Decimal('-123.45'),
                'blank': {},
                'extra': {
                    'deeper': 'more strings',
                    'nums': -12345678,
                },
            }
        }

        w = AMQPWriter()
        w.write_table(val)
        s = w.getvalue()

        r = AMQPReader(s)
        assert r.read_table() == val

    # -------------
    # Array

    def test_array_from_list(self):
        val = [1, 'foo', None]
        w = AMQPWriter()
        w.write_array(val)
        s = w.getvalue()

        assert_equal_binary(
            s, '\x00\x00\x00\x0EI\x00\x00\x00\x01S\x00\x00\x00\x03fooV',
        )

        r = AMQPReader(s)
        assert r.read_array() == val

    def test_array_from_tuple(self):
        val = (1, 'foo', None)
        w = AMQPWriter()
        w.write_array(val)
        s = w.getvalue()

        assert_equal_binary(
            s, '\x00\x00\x00\x0EI\x00\x00\x00\x01S\x00\x00\x00\x03fooV',
        )

        r = AMQPReader(s)
        assert r.read_array() == list(val)

    def test_table_with_array(self):
        val = {
            'foo': 7,
            'bar': Decimal('123345.1234'),
            'baz': 'this is some random string I typed',
            'blist': [1, 2, 3],
            'nlist': [1, [2, 3, 4]],
            'ndictl': {'nfoo': 8, 'nblist': [5, 6, 7]}
        }

        w = AMQPWriter()
        w.write_table(val)
        s = w.getvalue()

        r = AMQPReader(s)
        assert r.read_table() == val


class TestGenericContent:
    def test_generic_content_eq(self):
        msg_1 = GenericContent({'dummy': 'foo'})
        msg_2 = GenericContent({'dummy': 'foo'})
        msg_3 = GenericContent({'dummy': 'bar'})

        assert msg_1 == msg_1
        assert msg_1 == msg_2
        assert msg_1 != msg_3
        assert msg_1 is not None



from __future__ import absolute_import, division, print_function

__metaclass__ = type
import errno
import six
import socket
import ssl
from abc import ABCMeta, abstractmethod
import logging
from threading import RLock
from ssl import SSLError
import datetime
import time

from . import compat
from .proto import Frame
from .concurrency import synchronized
from .exceptions import UnexpectedFrame
from .utils import get_errno
from .spec import FrameType

log = logging.getLogger('amqpy')
compat.patch()

_UNAVAIL = {errno.EAGAIN, errno.EINTR, errno.ENOENT}

AMQP_PROTOCOL_HEADER = b'AMQP\x00\x00\x09\x01'  # bytes([65, 77, 81, 80, 0, 0, 9, 1])


class Transport:
    __metaclass__ = ABCMeta
    """Common superclass for TCP and SSL transports"""
    connected = False

    def __init__(self, host, port, connect_timeout, buf_size):
        """
        :param host: hostname or IP address
        :param port: port
        :param connect_timeout: connect timeout
        :type host: str
        :type port: int
        :type connect_timeout: float or None
        """
        self._rbuf = bytearray(buf_size)

        #: :type: datetime.datetime
        self.last_heartbeat_sent = None
        #: :type: datetime.datetime
        self.last_heartbeat_received = None

        self.last_heartbeat_sent_monotonic = 0.0

        # the purpose of the frame lock is to allow no more than one thread to read/write a frame
        # to the connection at any time
        self._frame_write_lock = RLock()
        self._frame_read_lock = RLock()

        self.sock = None

        # try to connect
        last_err = None
        for res in socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM, socket.IPPROTO_TCP):
            af, socktype, proto, canonname, sa = res
            try:
                self.sock = socket.socket(af, socktype, proto)
                self.sock.settimeout(connect_timeout)
                self.sock.connect(sa)
                break
            except socket.error as exc:
                self.sock.close()
                self.sock = None
                last_err = exc

        if not self.sock:
            # didn't connect, return the most recent error message
            raise socket.error(last_err)

        try:
            assert isinstance(self.sock, socket.socket)
            self.sock.settimeout(None)
            self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

            self._setup_transport()

            self.write(AMQP_PROTOCOL_HEADER)
        except (OSError, IOError, socket.error) as exc:
            if get_errno(exc) not in _UNAVAIL:
                self.connected = False
            raise

        self.connected = True

    def __del__(self):
        try:
            # socket module may have been collected by gc if this is called by a thread at shutdown
            if socket is not None:
                # noinspection PyBroadException
                try:
                    self.close()
                except:
                    pass
        finally:
            self.sock = None

    def _read(self, n, initial, _errnos):
        """Read from socket

        This is the default implementation. Subclasses may implement `read()` to simply call this
        method, or provide their own `read()` implementation.

        Note: According to SSL_read(3), it can at most return 16kB of data. Thus, we use an internal
        read buffer like to get the exact number of bytes wanted.

        Note: ssl.sock.read may cause ENOENT if the operation couldn't be performed (?).

        :param int n: exact number of bytes to read
        :return: data read
        :rtype: memoryview
        """
        mview = memoryview(self._rbuf)
        to_read = n
        while to_read:
            try:
                bytes_read = self.sock.recv_into(mview, to_read)
                mview = mview[bytes_read:]
                to_read -= bytes_read
            except socket.error as exc:
                if not initial and exc.errno in _errnos:
                    continue
                raise

            if not bytes_read:
                raise IOError('socket closed')

        return memoryview(self._rbuf)[:n]

    @abstractmethod
    def read(self, n, initial=False):
        """Read exactly `n` bytes from the peer

        :param n: number of bytes to read
        :type n: int
        :return: data read
        :rtype: bytes
        """
        pass

    @abstractmethod
    def write(self, s):
        """Completely write a string to the peer
        """

    def _setup_transport(self):
        """Do any additional initialization of the class (used by the subclasses)
        """
        pass

    def close(self):
        if self.sock is not None:
            # call shutdown first to make sure that pending messages reach the AMQP broker if the
            # program exits after calling this method
            self.sock.shutdown(socket.SHUT_RDWR)
            self.sock.close()
            self.sock = None
        self.connected = False

    @synchronized('_frame_read_lock')
    def read_frame(self):
        """Read frame from connection

        Note that the frame may be destined for any channel. It is permitted to interleave frames
        from different channels.

        :return: frame
        :rtype: amqpy.proto.Frame
        """
        frame = Frame()
        try:
            # read frame header: 7 bytes
            frame_header = self.read(7, True)
            frame.data.extend(frame_header)

            # read frame payload
            payload = self.read(frame.payload_size)
            frame.data.extend(payload)

            # read frame terminator byte
            frame_terminator = self.read(1)
            frame.data.extend(frame_terminator)

            if six.PY2:
                #: :type: int
                i_last_byte = six.byte2int(frame_terminator)
            else:
                # this fixes the change in memoryview in Python 3.3 (accessing an element returns the
                #  correct type)
                #: :type: int
                i_last_byte = six.byte2int(bytes(frame_terminator))
        except (OSError, IOError, socket.error) as exc:
            # don't disconnect for ssl read time outs (Python 3.2):
            # http://bugs.python.org/issue10272
            if isinstance(exc, SSLError) and 'timed out' in str(exc):
                raise socket.timeout()
            if get_errno(exc) not in _UNAVAIL and not isinstance(exc, socket.timeout):
                self.connected = False
            raise

        if i_last_byte == FrameType.END:
            if frame.frame_type == FrameType.HEARTBEAT:
                self.last_heartbeat_received = datetime.datetime.now()
            return frame
        else:
            raise UnexpectedFrame('Received {} while expecting 0xce (FrameType.END)'.format(hex(i_last_byte)))


    @synchronized('_frame_write_lock')
    def write_frame(self, frame):
        """Write frame to connection

        Note that the frame may be destined for any channel. It is permitted to interleave frames
        from different channels.

        :param frame: frame
        :type frame: amqpy.proto.Frame
        """
        try:
            self.write(frame.data)
        except socket.timeout:
            raise
        except (OSError, IOError, socket.error) as exc:
            if get_errno(exc) not in _UNAVAIL:
                self.connected = False
            raise

    def send_heartbeat(self):
        """Send a heartbeat to the server
        """
        self.last_heartbeat_sent = datetime.datetime.now()
        self.last_heartbeat_sent_monotonic = time.monotonic()
        self.write_frame(Frame(FrameType.HEARTBEAT))

    def is_alive(self):
        """Check if connection is alive

        This method is the primary way to check if the connection is alive.

        Side effects: This method may send a heartbeat as a last resort to check if the connection
        is alive.

        :return: True if connection is alive, else False
        :rtype: bool
        """
        if not self.sock:
            # we don't have a valid socket, this connection is definitely not alive
            return False

        if not self.connected:
            # the `transport` is not connected
            return False

        # recv with MSG_PEEK to check if the connection is alive
        # note: if there is data still in the buffer, this will not tell us anything
        # if hasattr(socket, 'MSG_PEEK') and not isinstance(self.sock, ssl.SSLSocket):
        #     prev = self.sock.gettimeout()
        #     self.sock.settimeout(0.0001)
        #     try:
        #         self.sock.recv(1, socket.MSG_PEEK)
        #     except socket.timeout:
        #         pass
        #     except socket.error:
        #         # the exception is usually (always?) a ConnectionResetError in Python 3.3+
        #         log.debug('socket.error, connection is closed')
        #         return False
        #     finally:
        #         self.sock.settimeout(prev)

        # send a heartbeat to check if the connection is alive
        try:
            self.send_heartbeat()
        except socket.error:
            return False

        return True


class SSLTransport(Transport):
    """Transport that works over SSL
    """

    def __init__(self, host, port, connect_timeout, frame_max, ssl_opts):
        self.ssl_opts = ssl_opts
        super(SSLTransport, self).__init__(host, port, connect_timeout, frame_max)

    def _setup_transport(self):
        """Wrap the socket in an SSL object
        """
        self.sock = ssl.wrap_socket(self.sock, **self.ssl_opts)

    def read(self, n, initial=False):
        """Read from socket

        According to SSL_read(3), it can at most return 16kb of data. Thus, we use an internal read
        buffer like `TCPTransport.read()` to get the exact number of bytes wanted.

        :param int n: exact number of bytes to read
        :return: data read
        :rtype: bytes
        """
        return self._read(n, initial, _errnos=(errno.ENOENT, errno.EAGAIN, errno.EINTR))

    def write(self, s):
        """Write a string out to the SSL socket fully
        """
        try:
            write = self.sock.write
        except AttributeError:
            # works around a bug in python socket library
            raise IOError('Socket closed')
        else:
            while s:
                n = write(s)
                if not n:
                    raise IOError('Socket closed')
                s = s[n:]


class TCPTransport(Transport):
    """Transport that deals directly with TCP socket
    """

    def read(self, n, initial=False):
        """Read exactly n bytes from the socket

        :param int n: exact number of bytes to read
        :return: data read
        :rtype: bytes
        """
        return self._read(n, initial, _errnos=(errno.EAGAIN, errno.EINTR))

    def write(self, s):
        self.sock.sendall(s)


def create_transport(host, port, connect_timeout, frame_max, ssl_opts=None):
    """Given a few parameters from the Connection constructor, select and create a subclass of
    Transport

    If `ssl_opts` is a dict, SSL will be used and `ssl_opts` will be passed to
    :func:`ssl.wrap_socket()`. In all other cases, SSL will not be used.

    :param host: host
    :param connect_timeout: connect timeout
    :param ssl_opts: ssl options passed to :func:`ssl.wrap_socket()`
    :type host: str
    :type connect_timeout: float or None
    :type ssl_opts: dict or None
    """
    if isinstance(ssl_opts, dict):
        return SSLTransport(host, port, connect_timeout, frame_max, ssl_opts)
    else:
        return TCPTransport(host, port, connect_timeout, frame_max)



"""Messages for AMQP
"""
from __future__ import absolute_import, division, print_function

__metaclass__ = type
import six
from . import spec
from amqpy.serialization import AMQPReader, AMQPWriter
import logging

log = logging.getLogger('amqpy')

__all__ = ['Message']


class GenericContent:
    """Base class for AMQP content

    Subclasses should override :attr:`PROPERTIES`.
    """
    __slots__ = ['properties']
    PROPERTIES = []

    def __init__(self, properties):
        """Save the properties appropriate to this AMQP content type in a 'properties' dictionary

        :param dict properties: content properties
        """
        #: Content properties
        #:
        #: :type: dict[str|unicode, str|dict]
        self.properties = properties

    def __eq__(self, other):
        """Check if this object has the same properties as another content object
        """
        return self.properties == other.properties

    def load_properties(self, raw_bytes):
        """Load raw bytes into :attr:`self.properties`

        The `raw_bytes` are the payload of a `FrameType.HEADER` frame, starting at a byte-offset
        of 12.
        """
        reader = AMQPReader(raw_bytes)

        # read 16-bit shorts until we get one with a low bit set to zero
        flags = []
        while True:
            flag_bits = reader.read_short()
            flags.append(flag_bits)
            if flag_bits & 1 == 0:
                break

        shift = 0
        d = {}
        flag_bits = None
        for prop_name, data_type in self.PROPERTIES:
            if shift == 0:
                if not flags:
                    break
                flag_bits, flags = flags[0], flags[1:]
                shift = 15
            if flag_bits & (1 << shift):
                d[prop_name] = getattr(reader, 'read_' + data_type)()
            shift -= 1

        self.properties = d

    def serialize_properties(self):
        """Serialize :attr:`self.properties` into raw bytes suitable to append
        to the payload of `FrameType.HEADER` frames
        """
        # write
        shift = 15
        flag_bits = 0
        flags = []
        prop_writer = AMQPWriter()
        for prop_name, data_type in self.PROPERTIES:
            val = self.properties.get(prop_name)
            if val is not None:
                if shift == 0:
                    flags.append(flag_bits)
                    flag_bits = 0
                    shift = 15

                flag_bits |= (1 << shift)
                if data_type != 'bit':
                    getattr(prop_writer, 'write_' + data_type)(val)
            shift -= 1
        flags.append(flag_bits)

        # write final data
        writer = AMQPWriter()
        for flag_bits in flags:
            writer.write_short(flag_bits)
        writer.write(prop_writer.getvalue())

        return writer.getvalue()


class Message(GenericContent):
    """A Message for use with the `Channel.basic_*` methods
    """
    __slots__ = ['body', 'channel', 'delivery_info']

    CLASS_ID = spec.Basic.CLASS_ID

    # instances of this class have these attributes, which are passed back and
    # forth as message properties between client and server
    PROPERTIES = [
        ('content_type', 'shortstr'),
        ('content_encoding', 'shortstr'),
        ('application_headers', 'table'),
        ('delivery_mode', 'octet'),
        ('priority', 'octet'),
        ('correlation_id', 'shortstr'),
        ('reply_to', 'shortstr'),
        ('expiration', 'shortstr'),
        ('message_id', 'shortstr'),
        ('timestamp', 'timestamp'),
        ('type', 'shortstr'),
        ('user_id', 'shortstr'),
        ('app_id', 'shortstr'),
        ('cluster_id', 'shortstr')
    ]

    def __init__(self, body='', channel=None, **properties):
        """
        If `body` is a `str`, then `content_encoding` will automatically be set to 'UTF-8', unless
        explicitly specified.

        Example::

            msg = Message('hello world', content_type='text/plain', application_headers={'foo': 7})

        :param body: message body
        :type body: bytes or str or unicode
        :param channel: associated channel
        :type channel: amqpy.channel.Channel
        :param properties:
            * content_type (shortstr): MIME content type
            * content_encoding (shortstr): MIME content encoding
            * application_headers: (table): Message header field table:
              dict[str, str|int|Decimal|datetime|dict]
            * delivery_mode: (octet): Non-persistent (1) or persistent (2)
            * priority (octet): The message priority, 0 to 9
            * correlation_id (shortstr) The application correlation identifier
            * reply_to (shortstr) The destination to reply to
            * expiration (shortstr): Message expiration specification
            * message_id (shortstr): The application message identifier
            * timestamp (datetime.datetime): The message timestamp
            * type (shortstr): The message type name
            * user_id (shortstr): The creating user id
            * app_id (shortstr): The creating application id
            * cluster_id (shortstr): Intra-cluster routing identifier
        """
        super(Message, self).__init__(properties)

        #: Message body (bytes or str or unicode)
        self.body = body

        #: Associated channel, set after receiving a message (amqpy.channel.Channel)
        self.channel = channel

        #: Delivery info, set after receiving a message (dict)
        self.delivery_info = {}

        if isinstance(body, six.string_types):
            # if the `body` is a string, automatically set the content_encoding
            # to UTF-8 if it hasn't already been set
            self.properties['content_encoding'] = properties.get('content_encoding', 'UTF-8')

    def __eq__(self, other):
        """Check if the properties and bodies of this Message and another Message are the same

        Received messages may contain a 'delivery_info' attribute, which isn't compared.
        """
        try:
            return super(Message, self).__eq__(other) and self.body == other.body
        except AttributeError:
            return False

    @property
    def application_headers(self):
        """Get application headers

        :return: application headers
        :rtype: dict
        """
        return self.properties.get('application_headers')

    @property
    def delivery_tag(self):
        """Get delivery tag

        :return: delivery tag
        :rtype: int
        """
        return self.delivery_info.get('delivery_tag')

    def ack(self):
        """Acknowledge message

        This is a convenience method which calls :meth:`self.channel.basic_ack()`
        """
        dt = self.delivery_tag
        if dt is not None:
            self.channel.basic_ack(dt)
        else:
            raise Exception('No delivery tag')

    def reject(self, requeue):
        """Reject message

        This is a convenience method which calls :meth:`self.channel.basic_reject()`

        :param bool requeue: requeue if True else discard the message
        """
        dt = self.delivery_tag
        if dt is not None:
            self.channel.basic_reject(dt, requeue)
        else:
            raise Exception('No delivery tag')



"""High-level representations of AMQP protocol objects
"""
from __future__ import absolute_import, division, print_function

__metaclass__ = type
import six
import struct
import logging

from .serialization import AMQPReader, AMQPWriter
from .spec import FrameType, method_t
from .message import Message

log = logging.getLogger('amqpy')


class Frame:
    """AMQP frame

    A `Frame` represents the lowest-level packet of data specified by the AMQP 0.9.1
    wire-level protocol. All methods and messages are packed into one or more frames before being
    sent to the peer.

    The format of the AMQP frame is as follows::

        offset:         0      1         3         7                 size+7      size+8
                        +------+---------+---------+-------------------+-----------+
                        | type | channel |  size   |  --- payload ---  | frame-end |
                        +------+---------+---------+-------------------+-----------+
        size (bytes)        1       2         4             size             1
    """

    __slots__ = ['data', '_frame_type', '_channel', '_payload_size']

    def __init__(self, frame_type=None, channel=0, payload=bytes()):
        """Create new Frame

        Leave all three parameters as default to create an empty frame whose `data` can be manually
        written to afterwards.

        :param frame_type: frame type
        :param channel: associated channel number
        :param payload: frame payload
        :type frame_type: int
        :type channel: int
        :type payload: bytes or bytearray
        """
        #: raw frame data; can be manually manipulated at any time
        #:
        #: :type: bytearray
        self.data = bytearray()

        self._frame_type = None
        self._channel = None
        self._payload_size = None

        # create bytearray from provided data
        if frame_type is not None:
            self._frame_type = frame_type
            self._channel = channel
            self._payload_size = len(payload)
            frame_format = '>BHI{}sB'.format(self._payload_size)
            self.data = struct.pack(frame_format, frame_type, channel, self._payload_size, payload,
                                    FrameType.END)

    @property
    def frame_type(self):
        """Get frame type

        :return: frame type
        :rtype: int
        """
        if self._frame_type is not None:
            return self._frame_type
        else:
            self._frame_type = struct.unpack_from('>B', self.data)[0]
            return self._frame_type

    @property
    def channel(self):
        """Get frame channel number

        :return: channel number
        :rtype: int
        """
        if self._channel is not None:
            return self._channel
        else:
            self._channel = struct.unpack_from('>H', self.data, 1)[0]
            return self._channel

    @property
    def payload_size(self):
        """Get frame payload size

        :return: payload size
        :rtype: int
        """
        if self._payload_size is not None:
            return self._payload_size
        else:
            self._payload_size = struct.unpack_from('>I', self.data, 3)[0]
            return self._payload_size

    @property
    def payload(self):
        """Get frame payload

        :return: payload
        :rtype: bytearray
        """
        return self.data[7:-1]


class Method:
    """AMQP method

    The AMQP 0.9.1 protocol specifies communication as sending and receiving "methods". Methods
    consist of a "class-id" and "method-id" and are represented by a `method_t` namedtuple in amqpy.
    Methods are packed into the payload of a `FrameType.METHOD` frame, and most methods can be fully
    sent in a single frame. If the method specified to be carrying content (such as a message), the
    method frame is followed by additional frames: a `FrameType.HEADER` frame, then zero or more
    `FrameType.BODY` frames.

    The format of the `FrameType.METHOD` frame's payload is as follows::

        offset:         0          2           4
                        +----------+-----------+-------------- - -
                        | class-id | method-id | arguments...
                        +----------+-----------+-------------- - -
        size (bytes):         2          2         variable

    The format of the `FrameType.HEADER` frame's payload is as follows::

        offset:         0          2        4           12               14
                        +----------+--------+-----------+----------------+------------------- - -
                        | class-id | weight | body size | property flags | property list...
                        +----------+--------+-----------+----------------+------------------- - -
        size (bytes):         2         2         8              2           variable

    The format of the `FrameType.BODY` frame's payload is simply raw binary data of the message
    body.
    """
    __slots__ = ['method_type', 'args', 'content', 'channel_id', '_body_bytes', '_expected_body_size']

    def __init__(self, method_type=None, args=None, content=None, channel_id=None):
        """
        :param method_type: method type
        :param args: method args
        :param content: content
        :param channel_id: the associated channel ID, if any
        :type method_type: method_t
        :type args: AMQPReader or AMQPWriter or None
        :type content: Message or None
        :type channel_id: int or None
        """
        #: :type: amqpy.spec.method_t
        self.method_type = method_type

        if isinstance(args, (AMQPReader, AMQPWriter)):
            self.args = args
        elif args is None:
            self.args = AMQPWriter()
        else:
            raise ValueError('args must be an instance of `AMQPReader` or `AMQPWriter`')

        #: :type: Message or None
        self.content = content  # Message if this method is carrying content
        #: :type: int
        self.channel_id = channel_id

        self._body_bytes = bytearray()  # used internally to store encoded GenericContent body
        self._expected_body_size = None  # set automatically when `load_header_frame()` is called

    def load_method_frame(self, frame):
        """Load method frame payload data

        This method is intended to be called when constructing a `Method` from incoming data.

        After calling, `self.method_type`, `self.args`, and `self.channel_id` will be loaded with
        data from the frame.

        :param frame: `FrameType.METHOD` frame
        :type frame: amqpy.proto.Frame
        """
        # noinspection PyTypeChecker
        self.method_type = method_t(*struct.unpack('>HH', frame.payload[:4]))
        self.args = AMQPReader(frame.payload[4:])
        self.channel_id = frame.channel

    def load_header_frame(self, frame):
        """Add header to partial method

        This method is intended to be called when constructing a `Method` from incoming data.

        :param frame: `FrameType.HEADER` frame
        :type frame: amqpy.proto.Frame
        """
        if not self.content:
            self.content = Message()

        # noinspection PyTypeChecker
        class_id, weight, self._expected_body_size = struct.unpack('>HHQ', frame.payload[:12])
        self.content.load_properties(frame.payload[12:])

    def load_body_frame(self, frame):
        """Add content to partial method

        This method is intended to be called when constructing a `Method` from incoming data.

        :param frame: `FrameType.BODY` frame
        :type frame: amqpy.proto.Frame
        """
        self._body_bytes.extend(frame.payload)
        if self.complete:
            self.content.body = bytes(self._body_bytes)

    @property
    def complete(self):
        """Check if the message that is carried by this method has been completely assembled,
        i.e. the expected number of bytes have been loaded

        This method is intended to be called when constructing a `Method` from incoming data.

        :return: True if method is complete, else False
        :rtype: bool
        """
        return self._expected_body_size == 0 or len(self._body_bytes) == self._expected_body_size

    def _pack_method(self):
        """Pack this method into a bytes object suitable for using as a payload for
        `FrameType.METHOD` frames

        This method is intended to be called when packing an already-completed `Method` into
        outgoing frames.

        :return: bytes
        :rtype: bytes
        """
        return struct.pack('>HH', self.method_type.class_id,
                           self.method_type.method_id) + self.args.getvalue()

    def _pack_header(self):
        """Pack this method into a bytes object suitable for using as a payload for
        `FrameType.HEADER` frames

        This method is intended to be called when packing an already-completed `Method` into
        outgoing frames.

        :return: bytes
        :rtype: bytes
        """
        if not self.content:
            raise ValueError('`_pack_header()` is only meaningful if there is content to pack')

        self._body_bytes = self.content.body
        if isinstance(self._body_bytes, six.string_types):
            # encode body to bytes
            coding = self.content.properties.setdefault('content_encoding', 'UTF-8')
            try:
                self._body_bytes = self.content.body.encode(coding)
            except LookupError:
                self._body_bytes = self.content.body

        properties = self.content.serialize_properties()
        return struct.pack('>HHQ', self.method_type.class_id, 0, len(self._body_bytes)) + properties

    def _pack_body(self, chunk_size):
        """Pack this method into a bytes object suitable for using as a payload for
        `FrameType.BODY` frames

        This method is intended to be called when packing an already-completed `Method` into
        outgoing frames.

        :param chunk_size: split up body into pieces that are at most `chunk_size` bytes each
        :type chunk_size: int
        :return: bytes generator
        :rtype: generator[bytes]
        """
        if not self.content:
            raise ValueError('`_pack_body()` is only meaningful if there is content to pack')

        for i in range(0, len(self._body_bytes), chunk_size):
            yield self._body_bytes[i:i + chunk_size]

    def dump_method_frame(self):
        """Create a method frame

        This method is intended to be called when sending frames for an already-completed `Method`.

        :return: `FrameType.METHOD` frame
        :rtype: amqpy.proto.Frame
        """
        frame = Frame(FrameType.METHOD, self.channel_id, self._pack_method())
        return frame

    def dump_header_frame(self):
        """Create a header frame

        This method is intended to be called when sending frames for an already-completed `Method`.

        :return: `FrameType.HEADER` frame
        :rtype: amqpy.proto.Frame
        """
        frame = Frame(FrameType.HEADER, self.channel_id, self._pack_header())
        return frame

    def dump_body_frame(self, chunk_size):
        """Create a body frame

        This method is intended to be called when sending frames for an already-completed `Method`.

        :param chunk_size: body chunk size in bytes; this is typically the maximum frame size - 8
        :type chunk_size: int
        :return: generator of `FrameType.BODY` frames
        :rtype: generator[amqpy.proto.Frame]
        """
        for payload in self._pack_body(chunk_size):
            frame = Frame(FrameType.BODY, self.channel_id, payload)
            yield frame



from __future__ import absolute_import, division, print_function

__metaclass__ = type
import six
import uuid
import logging
import sys

import pytest

from .. import Connection, Channel, Message, FrameSyntaxError, queue_declare_ok_t
from ..exceptions import AMQPError, ChannelError, PreconditionFailed, NotFound, AccessRefused
from .conftest import get_server_props

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout, style='{',
                    format='{asctime} {levelname:8} {message}',
                    datefmt='%Y/%m/%d %H:%M:%S')
log = logging.getLogger('amqpy')


class TestChannel:
    def test_encoding(self, ch, rand_rk):
        qname, _, _ = ch.queue_declare()
        ch.queue_bind(qname, 'amq.direct', routing_key=rand_rk)

        # no encoding, body passed through unchanged
        msg = Message('hello world')
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert isinstance(msg2.body, six.string_types)
        assert msg2.body == 'hello world'

        # default UTF-8 encoding of unicode body, returned as unicode
        msg = Message('hello world')
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'UTF-8'
        assert msg2.body == 'hello world'

        # explicit latin_1 encoding, still comes back as unicode
        msg = Message('hello world', content_encoding='latin_1')
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'latin_1'
        assert msg2.body == 'hello world'

        # plain string with specified encoding comes back as unicode
        msg = Message(six.u('hello w\xf6rld'), content_encoding='latin_1')
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'latin_1'
        assert msg2.body == six.u('hello w\u00f6rld')

        # plain string (bytes in Python 3.x) with bogus encoding
        # don't really care about latin_1, just want bytes
        test_bytes = six.u('hello w\xd6rld').encode('latin_1')
        msg = Message(test_bytes, content_encoding='I made this up')
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'I made this up'
        assert isinstance(msg2.body, bytes)
        assert msg2.body == test_bytes

        # turn off auto_decode for remaining tests
        ch.auto_decode = False

        # unicode body comes back as utf-8 encoded str
        msg = Message(six.u('hello w\u00f6rld'))
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'UTF-8'
        assert isinstance(msg2.body, bytes)
        assert msg2.body == six.u('hello w\xc3\xb6rld').encode('latin_1')

        # plain string with specified encoding stays plain string
        msg = Message(six.u('hello w\xf6rld'), content_encoding='latin_1')
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'latin_1'
        assert isinstance(msg2.body, bytes)
        assert msg2.body == six.u('hello w\xf6rld').encode('latin_1')

        # explicit latin_1 encoding, comes back as str
        msg = Message(six.u('hello w\u00f6rld'), content_encoding='latin_1')
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'latin_1'
        assert isinstance(msg2.body, bytes)
        assert msg2.body == six.u('hello w\xf6rld').encode('latin_1')

    def test_invalid_header(self, ch):
        """Test sending a message with an unserializable object in the header

        http://code.google.com/p/py-amqplib/issues/detail?id=17
        """
        qname, _, _ = ch.queue_declare()

        msg = Message(application_headers={'test': object()})

        with pytest.raises(FrameSyntaxError):
            ch.basic_publish(msg, routing_key=qname)


class TestChannelMode:
    def test_confirm_select(self, ch):
        ch.confirm_select()
        assert ch.mode == Channel.CH_MODE_CONFIRM

    def test_tx_select(self, ch):
        ch.tx_select()
        assert ch.mode == Channel.CH_MODE_TX

    def test_change_mode_raises(self, ch):
        ch.confirm_select()
        assert ch.mode == Channel.CH_MODE_CONFIRM

        with pytest.raises(PreconditionFailed):
            ch.tx_select()

        assert ch.mode == Channel.CH_MODE_NONE


class TestExchange:
    def test_exchange_declare_and_delete(self, ch):
        exch_name = 'test_exchange_{}'.format(uuid.uuid4())
        ch.exchange_declare(exch_name, 'direct')
        ch.exchange_delete(exch_name)

    def test_exchange_declare_passive_raises(self, ch):
        with pytest.raises(NotFound):
            exch_name = 'test_exchange_{}'.format(uuid.uuid4())
            ch.exchange_declare(exch_name, 'direct', passive=True)

    def test_exchange_redeclare_different_type_raises(self, ch, rand_exch):
        """Redeclaring an exchange with a different type should raise
        """
        ch.exchange_declare(rand_exch, 'direct')
        with pytest.raises(AMQPError):
            ch.exchange_declare(rand_exch, 'fanout')

    def test_exchange_redeclare_auto_delete_raises(self, ch, rand_exch):
        """Redeclaring an exchange with a different `auto_delete` should raise
        """
        ch.exchange_declare(rand_exch, 'direct', auto_delete=False)
        with pytest.raises(AMQPError):
            ch.exchange_declare(rand_exch, 'direct', auto_delete=True)

    def test_exchange_delete_nonexistent_raises(self, ch, rand_exch):
        """Test to ensure that deleting a nonexistent exchange raises `NotFound`

        Note: starting with RabbitMQ 3.2 (?), exchange.delete is an idempotent
        assertion that the exchange must not exist.

             We have made queue.delete into an idempotent assertion that the
             queue must not exist, in the same way that queue.declare asserts
             that it must. See https://www.rabbitmq.com/specification.html

        This means that the RabbitMQ server will not raise a 404 NOT FOUND
        channel exception when attempting to delete a nonexistent exchange.
        """
        server_props = get_server_props(ch.connection)

        if server_props[0] == 'RabbitMQ' and server_props[1] >= (3, 2, 0):
            ch.exchange_delete(rand_exch)
        else:
            with pytest.raises(NotFound):
                ch.exchange_delete(rand_exch)

    def test_exchange_delete_default(self, ch):
        with pytest.raises(AccessRefused):
            ch.exchange_delete('')

    def test_exchange_delete_in_use(self, ch, rand_exch, rand_queue, rand_rk):
        ch.exchange_declare(rand_exch, 'direct')
        ch.queue_declare(rand_queue)
        ch.queue_bind(rand_queue, rand_exch, rand_rk)

        with pytest.raises(PreconditionFailed):
            ch.exchange_delete(rand_exch, if_unused=True)

    def test_exchange_bind(self, ch):
        """Test exchange binding

        Network configuration is as follows (-> is forwards to):
        source_exchange -> dest_exchange -> queue.

        The test checks that once the message is published to the destination
        exchange(funtest.topic_dest) it is delivered to the queue.
        """
        test_routing_key = 'unit_test__key'
        dest_exchange = 'funtest.topic_dest_bind'
        source_exchange = 'funtest.topic_source_bind'

        ch.exchange_declare(dest_exchange, 'topic', auto_delete=True)
        ch.exchange_declare(source_exchange, 'topic', auto_delete=True)

        qname, _, _ = ch.queue_declare()
        ch.exchange_bind(dest_exch=dest_exchange, source_exch=source_exchange,
                         routing_key=test_routing_key)
        ch.queue_bind(qname, dest_exchange, routing_key=test_routing_key)

        msg = Message('funtest message', content_type='text/plain',
                      application_headers={'foo': 7, 'bar': 'baz'})

        ch.basic_publish(msg, source_exchange, routing_key=test_routing_key)

        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg == msg2

    def test_exchange_unbind(self, ch):
        dest_exchange = 'funtest.topic_dest_unbind'
        source_exchange = 'funtest.topic_source_unbind'
        test_routing_key = 'unit_test__key'

        ch.exchange_declare(dest_exchange, 'topic', auto_delete=True)
        ch.exchange_declare(source_exchange, 'topic', auto_delete=True)

        ch.exchange_bind(dest_exch=dest_exchange, source_exch=source_exchange,
                         routing_key=test_routing_key)

        ch.exchange_unbind(dest_exch=dest_exchange, source_exch=source_exchange,
                           routing_key=test_routing_key)


class TestQueue:
    def test_queue_declare(self, ch, rand_queue):
        ok = ch.queue_declare(rand_queue)
        assert isinstance(ok, queue_declare_ok_t)
        assert ok.queue == rand_queue
        assert ok.message_count == 0
        assert ok.consumer_count == 0

    def test_queue_bind_publish(self, ch, rand_rk):
        qname, _, _ = ch.queue_declare()
        ch.queue_bind(qname, 'amq.direct', routing_key=rand_rk)

        msg = Message('Hello, world!', content_type='text/plain',
                      application_headers={'foo': 7, 'bar': 'baz'})
        ch.basic_publish(msg, 'amq.direct', routing_key=rand_rk)

        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg == msg2

    def test_queue_delete_nonexistent(self, ch):
        """Test to ensure that deleting a nonexistent queue raises `NotFound`

        Note: starting with RabbitMQ 3.2 (?), queue.delete is an idempotent
        assertion that the queue must not exist.

            We have made queue.delete into an idempotent assertion that the
            queue must not exist, in the same way that queue.declare asserts
            that it must. See https://www.rabbitmq.com/specification.html

        This means that the RabbitMQ server will not raise a 404 NOT FOUND
        channel exception when attempting to delete a nonexistent queue.
        """
        server_props = get_server_props(ch.connection)
        if server_props[0] == 'RabbitMQ' and server_props[1] >= (3, 2, 0):
            assert ch.queue_delete('bogus_queue_that_does_not_exist') == 0
        else:
            with pytest.raises(NotFound):
                ch.queue_delete('bogus_queue_that_does_not_exist')

    def test_unbind(self, ch):
        my_routing_key = 'funtest.test_queue'

        qname, _, _ = ch.queue_declare()
        ch.queue_bind(qname, 'amq.direct', routing_key=my_routing_key)
        ch.queue_unbind(qname, 'amq.direct', routing_key=my_routing_key)

    def test_unbind_nonexistent(self, ch, rand_exch, rand_queue, rand_rk):
        """Test to ensure that unbinding a nonexistent binding raises `NotFound`

        Note: starting with RabbitMQ 3.2 (?), queue.delete is an idempotent
        assertion that the queue must not exist.

            We have made exchange.delete into an idempotent assertion that
            the exchange must not exist, in the same way that exchange.declare
            asserts that it must. See
            https://www.rabbitmq.com/specification.html

        This means that the RabbitMQ server will not raise a 404 NOT FOUND
        channel exception when attempting to delete a nonexistent queue.
        """
        server_props = get_server_props(ch.connection)

        if server_props[0] == 'RabbitMQ' and server_props[1] >= (3, 2, 0):
            ch.queue_unbind(rand_queue, rand_exch, rand_rk)
        else:
            with pytest.raises(NotFound):
                ch.queue_unbind(rand_queue, rand_exch, rand_rk)


class TestPublish:
    def test_publish(self, ch):
        ch.exchange_declare('funtest.fanout', 'fanout', auto_delete=True)
        msg = Message('funtest message', content_type='text/plain',
                      application_headers={'foo': 7, 'bar': 'baz'})
        ch.basic_publish(msg, 'funtest.fanout')

    def test_publish_large(self, ch):
        """Test sending some extra large messages
        """
        qname, _, _ = ch.queue_declare()

        for multiplier in [100, 1000, 10000, 50000, 100000, 500000]:
            msg = Message('this is a test message' * multiplier, content_type='text/plain',
                          application_headers={'foo': 7, 'bar': 'baz'})

            ch.basic_publish(msg, routing_key=qname)

            msg2 = ch.basic_get(no_ack=True)
            assert msg == msg2

    def test_publish_confirm(self, ch, rand_exch):
        queue_name = 'test.queue.publish'
        rk = queue_name
        ch.exchange_declare(rand_exch, 'direct', auto_delete=True)
        ch.queue_declare(queue_name)
        ch.queue_bind(queue_name, rand_exch, rk)
        msg = Message('funtest message', content_type='text/plain',
                      application_headers={'foo': 7, 'bar': 'baz'})
        ch.confirm_select()
        ch.basic_publish(msg, rand_exch, rk)

    def test_publish_tx(self, conn, rand_queue, rand_rk):
        """Transactions must work as expected
        """
        #: :type: amqpy.channel.Channel
        ch_a = conn.channel()
        ch_a.queue_declare(rand_queue)
        ch_a.queue_bind(rand_queue, 'amq.direct', rand_rk)
        # select transactional mode and publish a message
        ch_a.tx_select()
        ch_a.basic_publish(Message('Hello, world!'), 'amq.direct', rand_rk)
        ch_a.basic_publish(Message('Hello, world!'), 'amq.direct', rand_rk)

        # create a new channel, and check that there are no messages on the
        # queue
        #: :type: amqpy.channel.Channel
        ch_b = conn.channel()
        ok = ch_b.queue_declare(rand_queue, passive=True)
        assert ok.message_count == 0

        # now commit the transaction, then make sure there are two messages on
        # the queue
        ch_a.tx_commit()
        ok = ch_b.queue_declare(rand_queue, passive=True)
        assert ok.message_count == 2

    def test_basic_return(self, conn, rand_exch):
        assert isinstance(conn, Connection)
        ch = conn.channel()
        assert isinstance(ch, Channel)
        ch.exchange_declare(rand_exch, 'fanout')

        ch.confirm_select()

        msg = Message('funtest message', content_type='text/plain',
                      application_headers={'foo': 7, 'bar': 'baz'})

        ch.basic_publish(msg, rand_exch)
        assert ch.returned_messages.qsize() == 0
        ch.basic_publish(msg, rand_exch, mandatory=True)
        assert ch.returned_messages.qsize() == 1
        ch.basic_publish(msg, rand_exch, mandatory=True)
        assert ch.returned_messages.qsize() == 2
        ch.basic_publish(msg, rand_exch, mandatory=True)
        # 3 of the 4 messages we sent should have been returned
        assert ch.returned_messages.qsize() == 3
        log.debug('returned messages assertion ok')

    def test_defaults(self, ch):
        """Test how a queue defaults to being bound to an AMQP default
        exchange, and how publishing defaults to the default exchange, and
        basic_get defaults to getting from the most recently declared queue,
        and queue_delete defaults to deleting the most recently declared queue
        """
        msg = Message('funtest message', content_type='text/plain',
                      application_headers={'foo': 7, 'bar': 'baz'})

        qname, _, _ = ch.queue_declare()
        ch.basic_publish(msg, routing_key=qname)

        msg2 = ch.basic_get(no_ack=True)
        assert msg == msg2

        n = ch.queue_purge()
        assert n == 0

        n = ch.queue_delete()
        assert n == 0

    def test_survives_channel_error(self, ch):
        with pytest.raises(ChannelError):
            ch.queue_declare('krjqheewq_bogus', passive=True)
        ch.queue_declare('funtest_survive')
        ch.queue_declare('funtest_survive', passive=True)
        assert ch.queue_delete('funtest_survive') == 0



from __future__ import absolute_import, division, print_function

__metaclass__ = type
import gc
import os
from signal import SIGHUP
import threading
import time
import signal
from amqpy.login import login_response_plain

import pytest

from .. import Channel, NotFound, FrameError, spec, Connection
from ..proto import Method


class TestConnection:
    def test_create_channel(self, conn):
        ch = conn.channel(1)
        assert isinstance(ch, Channel)
        assert ch.channel_id == 1

        ch2 = conn.channel()
        assert ch2.channel_id != 1

        ch.close()
        ch2.close()

    def test_close(self, conn):
        """Make sure we've broken various references when closing channels and connections, to help
        with GC
        """
        # create a channel and make sure it's linked as we'd expect
        ch = conn.channel()
        assert 1 in conn.channels
        assert ch.connection == conn
        assert ch.is_open is True

        # close the channel and make sure the references are broken that we expect
        ch.close()
        assert ch.connection is None
        assert 1 not in conn.channels
        assert ch.callbacks == {}
        assert ch.is_open is False

        # close the connection and make sure the references we expect are gone
        conn.close()
        assert conn.connection is None
        assert conn.channels is not None

    def test_open_close_open(self):
        # create a new connection
        conn = Connection()

        # close the connection
        conn.close()

        # reopen the connection
        conn.connect()

    def test_is_alive(self, conn):
        assert conn.is_alive()

    def test_is_alive_after_close(self, conn):
        conn.close()
        assert conn.is_alive() is False

    def test_is_alive_chan_exception(self, conn, ch, rand_queue):
        """Make sure connection is still alive after a channel exception
        """
        with pytest.raises(NotFound):
            ch.queue_declare(rand_queue, passive=True)

        assert conn.is_alive()

    def test_is_alive_conn_exception(self, conn, rand_queue):
        """Make sure is_alive() returns False after a connection exception
        """
        ch = Channel(conn, 10)

        with pytest.raises(NotFound):
            ch.queue_declare(rand_queue, passive=True)

        with pytest.raises(FrameError):
            conn.method_writer.write_method(Method(spec.Queue.Declare, channel_id=10))
            conn.wait()

        assert conn.is_alive() is False

    def test_gc_closed(self, conn):
        """Make sure we've broken various references when closing channels and connections, to help
        with GC
        """
        unreachable_before = len(gc.garbage)
        # create a channel and make sure it's linked as we'd expect
        conn.channel()
        assert 1 in conn.channels

        # close the connection and make sure the references we expect are gone.
        conn.close()

        gc.collect()
        gc.collect()
        gc.collect()
        assert unreachable_before == len(gc.garbage)

    def test_gc_forget(self, conn):
        """Make sure the connection gets gc'ed when there is no more references to it
        """
        unreachable_before = len(gc.garbage)

        ch = conn.channel()
        assert 1 in conn.channels

        del ch

        gc.collect()
        gc.collect()
        gc.collect()
        assert unreachable_before == len(gc.garbage)

    @pytest.mark.skipif('sys.version_info >= (3, 5) or sys.version_info[0] == 2')
    def test_interrupted(self, conn):
        """Make sure to get InterruptedError if a read was interrupted
        """

        def sig_handler(sig, frame):
            pass

        signal.signal(SIGHUP, sig_handler)

        def interrupt_it():
            time.sleep(1)
            os.kill(os.getpid(), signal.SIGHUP)

        th = threading.Thread(target=interrupt_it)
        th.start()

        with pytest.raises(InterruptedError):
            conn.drain_events(2)


class TestLogin:
    def test_login_response_plain(self):
        b = login_response_plain('blah', 'blah')
        assert isinstance(b, bytes)



from __future__ import absolute_import, division, print_function

__metaclass__ = type
from threading import Lock
import sys
from collections import defaultdict, deque
import six
import logging
import socket
import errno
from .utils import get_errno

if six.PY2:
    from Queue import Queue
else:
    from queue import Queue

from .concurrency import synchronized
from .exceptions import UnexpectedFrame, Timeout, METHOD_NAME_MAP
from .proto import Method
from . import spec
from .spec import FrameType

log = logging.getLogger('amqpy')

__all__ = ['MethodReader']

# these received methods are followed by content headers and bodies
_CONTENT_METHODS = [
    spec.Basic.Return,
    spec.Basic.Deliver,
    spec.Basic.GetOk,
]


class MethodReader:
    """Read frames from the server and construct complete methods

    There should be one `MethodReader` instance per connection.

    In the case of a framing error, an :exc:`AMQPConnectionError` is placed in the queue.
    In the case of unexpected frames, an :exc:`ChannelError` is placed in the queue.
    """

    def __init__(self, transport):
        """
        :param transport: transport to read from
        :type transport: amqpy.transport.Transport
        """
        self.transport = transport
        self.sock = transport.sock

        # deque[Method or Exception]
        self.method_queue = deque()

        # dict[channel_id int: PartialMessage]
        self.partial_methods = {}

        # next expected frame type for each channel
        # dict[channel_id int: frame_type int]
        self.expected_types = defaultdict(lambda: FrameType.METHOD)

        self.frames_recv = 0  # total number of frames received

        self._method_read_lock = Lock()

    def _next_method(self):
        """Read the next method from the source and process it

        Once one complete method has been assembled, it is placed in the internal queue. This
        method will block until a complete `Method` has been constructed, which may consist of one
        or more frames.
        """
        while not self.method_queue:
            # keep reading frames until we have at least one complete method in the queue
            try:
                frame = self.transport.read_frame()
            except Exception as exc:
                # connection was closed? framing error?
                if six.PY2:
                    _, _, tb = sys.exc_info()
                    exc.tb = tb
                self.method_queue.append(exc)
                break

            self.frames_recv += 1

            if frame.frame_type not in (self.expected_types[frame.channel], 8):
                msg = 'Received frame type {} while expecting type: {}' \
                    .format(frame.frame_type, self.expected_types[frame.channel])
                self.method_queue.append(UnexpectedFrame(msg, channel_id=frame.channel))
            elif frame.frame_type == FrameType.METHOD:
                self._process_method_frame(frame)
            elif frame.frame_type == FrameType.HEADER:
                self._process_content_header(frame)
            elif frame.frame_type == FrameType.BODY:
                self._process_content_body(frame)

    def _process_method_frame(self, frame):
        """Process method frame

        :param frame: incoming frame
        :type frame: amqpy.proto.Frame
        """
        channel_id = frame.channel
        method = Method()
        method.load_method_frame(frame)

        if method.method_type in _CONTENT_METHODS:
            # save what we've got so far and wait for the content header frame
            self.partial_methods[channel_id] = method
            self.expected_types[channel_id] = spec.FrameType.HEADER
        else:
            # this is a complete method
            self.method_queue.append(method)

    def _process_content_header(self, frame):
        """Process Content Header frames

        :param frame: incoming frame
        :type frame: amqpy.proto.Frame
        """
        #: :type: amqpy.proto.Method
        method = self.partial_methods[frame.channel]
        method.load_header_frame(frame)

        if method.complete:
            # a bodyless message, we're done
            self.method_queue.append(method)
            self.partial_methods.pop(frame.channel, None)
            del self.expected_types[frame.channel]  # reset expected frame type for this channel
        else:
            # next expected frame type is FrameType.BODY
            self.expected_types[frame.channel] = spec.FrameType.BODY

    def _process_content_body(self, frame):
        """Process Content Body frames

        :param frame: incoming frame
        :type frame: amqpy.proto.Frame
        """
        method = self.partial_methods[frame.channel]
        method.load_body_frame(frame)

        if method.complete:
            # message is complete, append it to the queue
            self.method_queue.append(method)
            self.partial_methods.pop(frame.channel, None)
            del self.expected_types[frame.channel]  # reset expected frame type for this channel

    @synchronized('_method_read_lock')
    def _read_method(self):
        """Read a method from the peer

        :return: method
        :rtype: amqpy.proto.Method
        """
        # fully read and process next method
        self._next_method()
        method = self.method_queue.popleft()

        # `method` may sometimes be an `Exception`, raise it here
        if isinstance(method, Exception):
            raise method

        log.debug('{:7} channel: {} {} {}'
                  .format('Read:', method.channel_id,
                          method.method_type, METHOD_NAME_MAP[method.method_type]))
        return method

    def read_method(self, timeout=None):
        """Read method

        :param timeout: timeout
        :type timeout: float
        :return: method
        :rtype: amqpy.proto.Method
        :raise amqpy.exceptions.Timeout: if the operation times out
        """
        if timeout is None:
            return self._read_method()

        orig_timeout = self.sock.gettimeout()
        if orig_timeout != timeout:
            self.sock.settimeout(timeout)

        try:
            return self._read_method()
        except socket.timeout:
            raise Timeout()
        except socket.error as e:
            if get_errno(e) == errno.EAGAIN:
                raise Timeout()
            raise
        finally:
            if orig_timeout != timeout:
                self.sock.settimeout(orig_timeout)


class MethodWriter:
    """Write methods to the server by breaking them up and constructing multiple frames

    There should be one `MethodWriter` instance per connection, and all channels share that
    instance. This class is thread-safe. Any thread may call :meth:`write_method()` as long as no
    more than one thread is writing to any given `channel_id` at a time.
    """

    def __init__(self, transport, frame_max):
        """
        :param transport: transport to write to
        :param frame_max: maximum frame payload size in bytes
        :type transport: amqpy.transport.Transport
        :type frame_max: int
        """
        self.transport = transport
        self.frame_max = frame_max
        self.methods_sent = 0  # total number of methods sent

    def write_method(self, method):
        """Write method to connection, destined for the channel as set in `method.channel_id`

        This implementation uses a queue internally to prepare all frames before writing in order
        to detect issues.

        This method is thread safe only if the `channel_id` parameter is unique across concurrent
        invocations. The AMQP protocol allows interleaving frames destined for different channels,
        but not within the same channel. This means no more than one thread may safely operate on
        any given channel.

        :param method: method to write
        :type method: amqpy.proto.Method
        """
        transport = self.transport
        log.debug('{:7} channel: {} {} {}'
                  .format('Write:', method.channel_id,
                          method.method_type, METHOD_NAME_MAP[method.method_type]))
        frames = Queue()

        # construct a method frame
        frames.put(method.dump_method_frame())

        if method.content:
            # construct a header frame
            frames.put(method.dump_header_frame())

            # construct one or more body frames, which contain the body of the `Message`
            chunk_size = self.frame_max - 8
            for frame in method.dump_body_frame(chunk_size):
                frames.put(frame)

        while not frames.empty():
            transport.write_frame(frames.get())

        self.methods_sent += 1



from __future__ import absolute_import, division, print_function

__metaclass__ = type

import six
import io
from struct import pack

from .serialization import AMQPWriter

__all__ = ['login_responses']

def login_response_amqplain(userid, password):
    response = AMQPWriter()
    response.write_table({'LOGIN': userid, 'PASSWORD': password})
    return response.getvalue()[4:]  # skip the length

def login_response_plain(userid, password):
    """https://tools.ietf.org/html/rfc4616"""
    response = io.BytesIO()
    if isinstance(userid, six.string_types):
        userid = userid.encode('utf-8')
    if isinstance(password, six.string_types):
        password = password.encode('utf-8')
    # authzid
    response.write(userid)
    response.write(pack('B', 0))
    # authcid
    response.write(userid)
    response.write(pack('B', 0))
    # passwd
    response.write(password)
    return response.getvalue()

login_responses = {
    'AMQPLAIN': login_response_amqplain,
    'PLAIN': login_response_plain,
}



from __future__ import absolute_import, division, print_function

__metaclass__ = type


def get_errno(exc):
    try:
        return exc.errno
    except AttributeError:
        try:
            # e.args = (errno, reason)
            if isinstance(exc.args, tuple) and len(exc.args) == 2:
                return exc.args[0]
        except AttributeError:
            pass
    return 0



from __future__ import absolute_import, division, print_function

__metaclass__ = type
import logging
import time
from functools import wraps

from . import compat

compat.patch()  # monkey-patch time.perf_counter

log = logging.getLogger('amqpy')


def synchronized(lock_name):
    """Decorator for automatically acquiring and releasing lock for method call

    This decorator accesses the `lock_name` :class:`threading.Lock` attribute of the instance that
    the wrapped method is bound to. The lock is acquired (blocks indefinitely) before the method is
    called. After the method has executed, the lock is released.

    Decorated methods should not be long-running operations, since the lock is held for the duration
    of the method's execution.

    :param lock_name: name of :class:`threading.Lock` object
    """

    def decorator(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            lock = getattr(self, lock_name)
            acquired = lock.acquire(False)
            if not acquired:
                # log.debug('> Wait to acquire lock for [{}]'.format(f.__qualname__))
                start_time = time.perf_counter()
                lock.acquire()
                tot_time = time.perf_counter() - start_time
                if tot_time > 5:
                    # only log if waited for more than 10s to acquire lock
                    log.warn('Acquired lock for [{}] in: {:.3f}s'.format(f.__qualname__, tot_time))
            try:
                retval = f(self, *args, **kwargs)
            finally:
                lock.release()
            return retval

        return wrapper

    return decorator


def synchronized_connection():
    """Decorator for automatically acquiring and releasing a connection-level
    lock for method call
    """

    def decorator(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            try:
                lock = self.conn_lock
            except AttributeError:
                lock = self.connection.conn_lock
            acquired = lock.acquire(False)
            if not acquired:
                # log.debug('> Wait to acquire lock for [{}]'.format(f.__qualname__))
                start_time = time.perf_counter()
                lock.acquire()
                tot_time = time.perf_counter() - start_time
                if tot_time > 5:
                    # only log if waited for more than 10s to acquire lock
                    log.warn('Acquired lock for [{}] in: {:.3f}s'.format(f.__qualname__, tot_time))
            try:
                retval = f(self, *args, **kwargs)
            finally:
                lock.release()
            return retval

        return wrapper

    return decorator



"""AMQP Channels
"""
from __future__ import absolute_import, division, print_function

__metaclass__ = type
import logging
import six

if six.PY2:
    from Queue import Queue
else:
    from queue import Queue

from .proto import Method
from .concurrency import synchronized_connection
from .abstract_channel import AbstractChannel
from .exceptions import ChannelError, ConsumerCancelled, error_for_code
from .spec import basic_return_t, queue_declare_ok_t, method_t
from .serialization import AMQPWriter
from . import spec

__all__ = ['Channel']

log = logging.getLogger('amqpy')


class Channel(AbstractChannel):
    """
    The channel class provides methods for a client to establish and operate an AMQP channel. All
    public members are fully thread-safe.
    """
    ### constants
    #: Default channel mode
    CH_MODE_NONE = 0
    #: Transaction mode
    CH_MODE_TX = 1
    #: Publisher confirm mode (RabbitMQ extension)
    CH_MODE_CONFIRM = 2

    def __init__(self, connection, channel_id=None, auto_decode=True):
        """Create a channel bound to a connection and using the specified numeric channel_id, and
        open on the server

        If `auto_decode` is enabled (default), incoming Message bodies will be automatically decoded
        to `str` if possible.

        :param connection: the channel's associated Connection
        :param channel_id: the channel's assigned channel ID
        :param auto_decode: enable auto decoding of message bodies
        :type connection: amqpy.connection.Connection
        :type channel_id: int or None
        :type auto_decode: bool
        """
        if channel_id:
            # noinspection PyProtectedMember
            connection._claim_channel_id(channel_id)
        else:
            # noinspection PyProtectedMember
            channel_id = connection._get_free_channel_id()

        super(Channel, self).__init__(connection, channel_id)

        # auto decode received messages
        self.auto_decode = auto_decode

        ### channel state variables:

        #: Current channel open/closed state
        #:
        #: :type: bool
        self.is_open = False

        #: Current channel active state (flow control)
        #:
        #: :type: bool
        self.active = True

        #: Channel mode state (default, transactional, publisher confirm)
        #:
        #: :type: int
        self.mode = 0

        #: Returned messages that the server was unable to deliver
        #:
        #: :type: queue.Queue
        self.returned_messages = Queue()

        # consumer callbacks dict[consumer_tag str: callable]
        self.callbacks = {}

        # consumer cancel callbacks dict dict[consumer_tag str: callable]
        self.cancel_callbacks = {}

        # set of consumers that have opted for `no_ack` delivery (server will not expect an ack
        # for delivered messages)
        self.no_ack_consumers = set()

        # open the channel
        self._open()

    def _close(self):
        """Tear down this object, after we've agreed to close with the server
        """
        log.debug('Channel close #{}'.format(self.channel_id))
        self.is_open = False
        channel_id, self.channel_id = self.channel_id, None
        connection, self.connection = self.connection, None
        if connection:
            connection.channels.pop(channel_id, None)
            # noinspection PyProtectedMember
            connection._avail_channel_ids.append(channel_id)
        self.callbacks.clear()
        self.cancel_callbacks.clear()
        self.no_ack_consumers.clear()

    def _open(self):
        """Open the channel
        """
        if self.is_open:
            return

        self._send_open()

    def _revive(self):
        self.is_open = False
        self.mode = self.CH_MODE_NONE
        self._send_open()

    @synchronized_connection()
    def close(self, reply_code=0, reply_text='', method_type=method_t(0, 0)):
        """Request a channel close

        This method indicates that the sender wants to close the channel. This may be due to
        internal conditions (e.g. a forced shut-down) or due to an error handling a specific method,
        i.e. an exception When a close is due to an exception, the sender provides the class and
        method id of the method which caused the exception.

        :param reply_code: the reply code
        :param reply_text: localized reply text
        :param method_type: if close is triggered by a failing method, this is the method that
            caused it
        :type reply_code: int
        :type reply_text: str
        :type method_type: amqpy.spec.method_t
        """
        try:
            if not self.is_open or self.connection is None:
                return

            args = AMQPWriter()
            args.write_short(reply_code)
            args.write_shortstr(reply_text)
            args.write_short(method_type.class_id)
            args.write_short(method_type.method_id)
            self._send_method(Method(spec.Channel.Close, args))
            return self.wait_any([spec.Channel.Close, spec.Channel.CloseOk])
        finally:
            self.connection = None

    def _cb_close(self, method):
        """Respond to a channel close sent by the server

        This method indicates that the sender (server) wants to close the channel. This may be due
        to internal conditions (e.g. a forced shut-down) or due to an error handling a specific
        method, i.e. an exception. When a close is due to an exception, the sender provides the
        class and method id of the method which caused the exception.

        This method sends a "close-ok" to the server, then re-opens the channel.
        """
        args = method.args
        reply_code = args.read_short()
        reply_text = args.read_shortstr()
        class_id = args.read_short()
        method_id = args.read_short()

        self._send_method(Method(spec.Channel.CloseOk))
        self.is_open = False

        # re-open the channel
        self._revive()

        # get information about the method which caused the server to close the channel
        method_type = method_t(class_id, method_id)
        raise error_for_code(reply_code, reply_text, method_type, ChannelError, self.channel_id)

    def _cb_close_ok(self, method):
        """Confirm a channel close

        This method confirms a Channel.Close method and tells the recipient that it is safe to
        release resources for the channel and close the socket.
        """
        assert method
        self._close()

    @synchronized_connection()
    def flow(self, active):
        """Enable/disable flow from peer

        This method asks the peer to pause or restart the flow of content data. This is a simple
        flow-control mechanism that a peer can use to avoid overflowing its queues or otherwise
        finding itself receiving more messages than it can process. Note that this method is not
        intended for window control The peer that receives a request to stop sending content
        should finish sending the current content, if any, and then wait until it receives a Flow
        restart method.

        :param active: True: peer starts sending content frames; False: peer stops sending content
            frames
        :type active: bool
        """
        args = AMQPWriter()
        args.write_bit(active)
        self._send_method(Method(spec.Channel.Flow, args))
        return self.wait_any([spec.Channel.FlowOk, self._cb_flow_ok])

    def _cb_flow(self, method):
        """Enable/disable flow from peer

        This method asks the peer to pause or restart the flow of content data. This is a simple
        flow-control mechanism that a peer can use to avoid overflowing its queues or otherwise
        finding itself receiving more messages than it can process. Note that this method is not
        intended for window control The peer that receives a request to stop sending content
        should finish sending the current content, if any, and then wait until it receives a Flow
        restart method.
        """
        args = method.args
        self.active = args.read_bit()
        self._send_flow_ok(self.active)

    def _send_flow_ok(self, active):
        """Confirm a flow method

        Confirms to the peer that a flow command was received and processed.

        :param active: True: peer starts sending content frames; False: peer stops sending content
            frames
        :type active: bool
        """
        args = AMQPWriter()
        args.write_bit(active)
        self._send_method(Method(spec.Channel.FlowOk, args))

    def _cb_flow_ok(self, method):
        """Confirm a flow method

        Confirms to the peer that a flow command was received and processed.
        """
        args = method.args
        return args.read_bit()

    def _send_open(self):
        """Open a channel

        This method opens a channel.
        """
        args = AMQPWriter()
        args.write_shortstr('')  # reserved
        self._send_method(Method(spec.Channel.Open, args))
        return self.wait(spec.Channel.OpenOk)

    def _cb_open_ok(self, method):
        """Handle received "open-ok"

        The server sends this method to signal to the client that this channel is ready for use.
        """
        assert method
        self.is_open = True
        log.debug('Channel open')

    @synchronized_connection()
    def exchange_declare(self, exchange, exch_type, passive=False, durable=False, auto_delete=True,
                         nowait=False, arguments=None):
        """Declare exchange, create if needed

        * Exchanges cannot be redeclared with different types. The client MUST not attempt to
          redeclare an existing exchange with a different type than used in the original
          Exchange.Declare method.
        * This method creates an exchange if it does not already exist, and if the exchange
          exists, verifies that it is of the correct and expected class.
        * The server must ignore the `durable` field if the exchange already exists.
        * The server must ignore the `auto_delete` field if the exchange already exists.
        * If `nowait` is enabled and the server could not complete the method, it will raise a
          channel or connection exception.
        * `arguments` is ignored if passive is True.

        :param str exchange: exchange name
        :param str exch_type: exchange type (direct, fanout, etc.)
        :param bool passive: do not create exchange; client can use this to check whether an
            exchange exists
        :param bool durable: mark exchange as durable (remain active after server restarts)
        :param bool auto_delete: auto-delete exchange when all queues have finished using it
        :param bool nowait: if set, the server will not respond to the method and the client should
            not wait for a reply
        :param dict arguments: exchange declare arguments
        :raise AccessRefused: if attempting to declare an exchange with a reserved name (amq.*)
        :raise NotFound: if `passive` is enabled and the exchange does not exist
        :return: None
        """
        arguments = arguments or {}
        args = AMQPWriter()
        args.write_short(0)  # reserved-1
        args.write_shortstr(exchange)  # exchange name
        args.write_shortstr(exch_type)  # exchange type
        args.write_bit(passive)  # passive
        args.write_bit(durable)  # durable
        args.write_bit(auto_delete)  # auto-delete
        args.write_bit(False)  # internal
        args.write_bit(nowait)
        args.write_table(arguments)
        self._send_method(Method(spec.Exchange.Declare, args))

        if not nowait:
            return self.wait(spec.Exchange.DeclareOk)

    def _cb_exchange_declare_ok(self, method):
        """Confirms an exchange declaration

        The server sends this method to confirm a Declare method and confirms the name of the
        exchange, essential for automatically-named exchanges.
        """
        pass

    @synchronized_connection()
    def exchange_delete(self, exchange, if_unused=False, nowait=False):
        """Delete an exchange

        This method deletes an exchange.

        * If the exchange does not exist, the server must raise a channel exception. When an
          exchange is deleted, all queue bindings on the exchange are cancelled.
        * If `if_unused` is set, and the exchange has queue bindings, the server must raise a
          channel exception.

        :param str exchange: exchange name
        :param bool if_unused: delete only if unused (has no queue bindings)
        :param bool nowait: if set, the server will not respond to the method and the client should
            not wait for a reply
        :raise NotFound: if exchange with `exchange` does not exist
        :raise PreconditionFailed: if attempting to delete a queue with bindings and `if_unused` is
            set
        :return: None
        """
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(exchange)
        args.write_bit(if_unused)
        args.write_bit(nowait)
        self._send_method(Method(spec.Exchange.Delete, args))

        if not nowait:
            return self.wait(spec.Exchange.DeleteOk)

    def _cb_exchange_delete_ok(self, method):
        """Confirm deletion of an exchange

        The server sends this method to confirm that the deletion of an exchange was successful.
        """
        pass

    @synchronized_connection()
    def exchange_bind(self, dest_exch, source_exch='', routing_key='', nowait=False,
                      arguments=None):
        """Bind an exchange to an exchange

        * Both the `dest_exch` and `source_exch` must already exist. Blank exchange names mean
          the default exchange.
        * A server MUST allow and ignore duplicate bindings - that is, two or more bind methods
          for a specific exchanges, with identical arguments - without treating these as an error.
        * A server MUST allow cycles of exchange bindings to be created including allowing an
          exchange to be bound to itself.
        * A server MUST not deliver the same message more than once to a destination exchange,
          even if the topology of exchanges and bindings results in multiple (even infinite)
          routes to that exchange.

        :param str dest_exch: name of destination exchange to bind
        :param str source_exch: name of source exchange to bind
        :param str routing_key: routing key for the binding (note: not all exchanges use a
            routing key)
        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        :param dict arguments: binding arguments, specific to the exchange class
        """
        arguments = {} if arguments is None else arguments
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(dest_exch)
        args.write_shortstr(source_exch)
        args.write_shortstr(routing_key)
        args.write_bit(nowait)
        args.write_table(arguments)
        self._send_method(Method(spec.Exchange.Bind, args))

        if not nowait:
            return self.wait(spec.Exchange.BindOk)

    @synchronized_connection()
    def exchange_unbind(self, dest_exch, source_exch='', routing_key='', nowait=False,
                        arguments=None):
        """Unbind an exchange from an exchange

        * If the unbind fails, the server must raise a connection exception. The server must not
            attempt to unbind an exchange that does not exist from an exchange.
        * Blank exchange names mean the default exchange.

        :param str dest_exch: destination exchange name
        :param str source_exch: source exchange name
        :param str routing_key: routing key to unbind
        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        :param dict arguments: binding arguments, specific to the exchange class
        """
        arguments = {} if arguments is None else arguments
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(dest_exch)
        args.write_shortstr(source_exch)
        args.write_shortstr(routing_key)
        args.write_bit(nowait)
        args.write_table(arguments)
        self._send_method(Method(spec.Exchange.Unbind, args))

        if not nowait:
            return self.wait(spec.Exchange.UnbindOk)

    def _cb_exchange_bind_ok(self, method):
        """Confirm bind successful

        The server sends this method to confirm that the bind was successful.
        """
        pass

    def _cb_exchange_unbind_ok(self, method):
        """Confirm unbind successful

        The server sends this method to confirm that the unbind was successful.
        """
        pass

    @synchronized_connection()
    def queue_bind(self, queue, exchange='', routing_key='', nowait=False, arguments=None):
        """Bind queue to an exchange

        This method binds a queue to an exchange. Until a queue is bound it will not receive any
        messages. In a classic messaging model, store-and-forward queues are bound to a dest
        exchange and subscription queues are bound to a dest_wild exchange.

        * The server must allow and ignore duplicate bindings without treating these as an error.
        * If a bind fails, the server must raise a connection exception.
        * The server must not allow a durable queue to bind to a transient exchange. If a client
          attempts this, the server must raise a channel exception.
        * The server should support at least 4 bindings per queue, and ideally, impose no limit
          except as defined by available resources.

        * If the client did not previously declare a queue, and the `queue` is empty, the server
          must raise a connection exception with reply code 530 (not allowed).
        * If `queue` does not exist, the server must raise a channel exception with reply code
          404 (not found).
        * If `exchange` does not exist, the server must raise a channel exception with reply code
          404 (not found).

        :param str queue: name of queue to bind; blank refers to the last declared queue for this
            channel
        :param str exchange: name of exchange to bind to
        :param str routing_key: routing key for the binding
        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        :param dict arguments: binding arguments, specific to the exchange class
        """
        arguments = {} if arguments is None else arguments
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(queue)
        args.write_shortstr(exchange)
        args.write_shortstr(routing_key)
        args.write_bit(nowait)
        args.write_table(arguments)
        self._send_method(Method(spec.Queue.Bind, args))

        if not nowait:
            return self.wait(spec.Queue.BindOk)

    def _cb_queue_bind_ok(self, method):
        """Confirm bind successful

        The server sends this method to confirm that the bind was successful.
        """
        pass

    @synchronized_connection()
    def queue_unbind(self, queue, exchange, routing_key='', nowait=False, arguments=None):
        """Unbind a queue from an exchange

        This method unbinds a queue from an exchange.

        * If a unbind fails, the server MUST raise a connection exception.
        * The client must not attempt to unbind a queue that does not exist.
        * The client must not attempt to unbind a queue from an exchange that does not exist.

        :param str queue: name of queue to unbind, leave blank to refer to the last declared
            queue on this channel
        :param str exchange: name of exchange to unbind, leave blank to refer to default exchange
        :param str routing_key: routing key of binding
        :param dict arguments: binding arguments, specific to the exchange class
        """
        arguments = {} if arguments is None else arguments
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(queue)
        args.write_shortstr(exchange)
        args.write_shortstr(routing_key)
        # args.write_bit(nowait)
        args.write_table(arguments)
        self._send_method(Method(spec.Queue.Unbind, args))

        if not nowait:
            return self.wait(spec.Queue.UnbindOk)

    def _cb_queue_unbind_ok(self, method):
        """Confirm unbind successful

        This method confirms that the unbind was successful.
        """
        pass

    @synchronized_connection()
    def queue_declare(self, queue='', passive=False, durable=False, exclusive=False,
                      auto_delete=True, nowait=False,
                      arguments=None):
        """Declare queue, create if needed

        This method creates or checks a queue. When creating a new queue the client can specify
        various properties that control the durability of the queue and its contents, and the level
        of sharing for the queue. A tuple containing the queue name, message count, and consumer
        count is returned, which is essential for declaring automatically named queues.

        * If `passive` is specified, the server state is not modified (a queue will not be
          declared), and the server only checks if the specified queue exists and returns its
          properties. If the queue does not exist, the server must raise a 404 NOT FOUND channel
          exception.
        * The server must create a default binding for a newly-created queue to the default
          exchange, which is an exchange of type 'direct'.
        * Queue names starting with 'amq.' are reserved for use by the server. If an attempt is
          made to declare a queue with such a name, and the `passive` flag is disabled, the server
          must raise a 403 ACCESS REFUSED connection exception.
        * The server must raise a 405 RESOURCE LOCKED channel exception if an attempt is made to
          access a queue declared as exclusive by another open connection.
        * The server must ignore the `auto_delete` flag if the queue already exists.

        RabbitMQ supports the following useful additional arguments:

        * x-max-length (int): maximum queue size
            * Queue length is a measure that takes into account ready messages, ignoring
              unacknowledged messages and message size. Messages will be dropped or dead-lettered
              from the front of the queue to make room for new messages once the limit is reached.

        :param str queue: queue name; leave blank to let the server generate a name automatically
        :param bool passive: do not create queue; client can use this to check whether a queue
            exists
        :param bool durable: mark as durable (remain active after server restarts)
        :param bool exclusive: mark as exclusive (can only be consumed from by this connection);
            implies `auto_delete`
        :param bool auto_delete: auto-delete queue when all consumers have finished using it
        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        :param dict arguments: exchange declare arguments
        :raise NotFound: if `passive` is enabled and the queue does not exist
        :raise AccessRefused: if an attempt is made to declare a queue with a reserved name
        :raise ResourceLocked: if an attempt is made to access an exclusive queue declared by
            another open connection
        :return: queue_declare_ok_t(queue, message_count, consumer_count), or None if `nowait`
        :rtype: queue_declare_ok_t or None
        """
        arguments = arguments or {}
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(queue)
        args.write_bit(passive)
        args.write_bit(durable)
        args.write_bit(exclusive)
        args.write_bit(auto_delete)
        args.write_bit(nowait)
        args.write_table(arguments)
        self._send_method(Method(spec.Queue.Declare, args))

        if not nowait:
            return self.wait(spec.Queue.DeclareOk)

    def _cb_queue_declare_ok(self, method):
        """Confirm a queue declare

        This method is called when the server responds to a `queue.declare`.

        :return: queue_declare_ok_t(queue, message_count, consumer_count), or None if `nowait`
        :rtype: queue_declare_ok_t or None
        """
        args = method.args
        return queue_declare_ok_t(args.read_shortstr(), args.read_long(), args.read_long())

    @synchronized_connection()
    def queue_delete(self, queue='', if_unused=False, if_empty=False, nowait=False):
        """Delete a queue

        This method deletes a queue. When a queue is deleted any pending messages are sent to a
        dead-letter queue if this is defined in the server configuration, and all consumers on the
        queue are cancelled.

        :param str queue: name of queue to delete, empty string refers to last declared queue on
            this channel
        :param bool if_unused: delete only if unused (has no consumers); raise a channel
            exception otherwise
        :param bool if_empty: delete only if empty; raise a channel exception otherwise
        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        :raise NotFound: if `queue` does not exist
        :raise PreconditionFailed: if `if_unused` or `if_empty` conditions are not met
        :return: number of messages deleted
        :rtype: int
        """
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(queue)
        args.write_bit(if_unused)
        args.write_bit(if_empty)
        args.write_bit(nowait)
        self._send_method(Method(spec.Queue.Delete, args))

        if not nowait:
            return self.wait(spec.Queue.DeleteOk)

    def _cb_queue_delete_ok(self, method):
        """Confirm deletion of a queue

        This method confirms the deletion of a queue.

        PARAMETERS:
            message_count: long

                number of messages purged

                Reports the number of messages purged.
        """
        args = method.args
        return args.read_long()

    @synchronized_connection()
    def queue_purge(self, queue='', nowait=False):
        """Purge a queue

        This method removes all messages from a queue. It does not cancel consumers. Purged messages
        are deleted without any formal "undo" mechanism.

        * On transacted channels the server MUST not purge messages that have already been sent
          to a client but not yet acknowledged.
        * If nowait is False, this method returns a message count.

        :param str queue: queue name to purge; leave blank to refer to last declared queue for
            this channel
        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        :return: message count (if nowait is False)
        :rtype: int or None
        """
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(queue)
        args.write_bit(nowait)
        self._send_method(Method(spec.Queue.Purge, args))

        if not nowait:
            return self.wait(spec.Queue.PurgeOk)

    def _cb_queue_purge_ok(self, method):
        """Confirms a queue purge

        This method confirms the purge of a queue.

        PARAMETERS:
            message_count: long

                number of messages purged

                Reports the number of messages purged.
        """
        args = method.args
        return args.read_long()

    @synchronized_connection()
    def basic_ack(self, delivery_tag, multiple=False):
        """Acknowledge one or more messages

        This method acknowledges one or more messages delivered via the Deliver or Get-Ok methods.
        The client can ask to confirm a single message or a set of messages up to and including a
        specific message.

        * The delivery tag is valid only within the same channel that the message was received.
        * Set `delivery_tag` to `0` and `multiple` to `True` to acknowledge all outstanding
          messages.
        * If the `delivery_tag` is invalid, the server must raise a channel exception.

        :param int delivery_tag: server-assigned delivery tag; 0 means "all messages received so
            far"
        :param bool multiple: if set, the `delivery_tag` is treated as "all messages up to and
            including"
        """
        args = AMQPWriter()
        args.write_longlong(delivery_tag)
        args.write_bit(multiple)
        self._send_method(Method(spec.Basic.Ack, args))

    @synchronized_connection()
    def basic_cancel(self, consumer_tag, nowait=False):
        """End a queue consumer

        This method cancels a consumer. This does not affect already delivered messages, but it does
        mean the server will not send any more messages for that consumer. The client may receive an
        arbitrary number of messages in between sending the cancel method and receiving the
        cancel-ok reply.

        * If the queue no longer exists when the client sends a cancel command, or the consumer
          has been cancelled for other reasons, this command has no effect.

        :param str consumer_tag: consumer tag, valid only within the current connection and channel
        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        """
        if self.connection is not None:
            self.no_ack_consumers.discard(consumer_tag)
            args = AMQPWriter()
            args.write_shortstr(consumer_tag)
            args.write_bit(nowait)
            self._send_method(Method(spec.Basic.Cancel, args))
            return self.wait(spec.Basic.CancelOk)

    def _cb_basic_cancel_notify(self, method):
        """Consumer cancelled by server.

        Most likely the queue was deleted.
        """
        args = method.args
        consumer_tag = args.read_shortstr()
        callback = self._on_cancel(consumer_tag)
        if callback:
            callback(consumer_tag)
        else:
            raise ConsumerCancelled(consumer_tag, spec.Basic.Cancel)

    def _cb_basic_cancel_ok(self, method):
        """Confirm a cancelled consumer

        This method confirms that the cancellation was completed.

        PARAMETERS: consumer_tag: shortstr

                consumer tag

                Identifier for the consumer, valid within the current connection.

                RULE:

                    The consumer tag is valid only within the channel from which the consumer was
                    created. I.e. a client
                    MUST NOT create a consumer in one channel and then use it in another.
        """
        args = method.args
        consumer_tag = args.read_shortstr()
        self._on_cancel(consumer_tag)

    def _on_cancel(self, consumer_tag):
        """
        :param consumer_tag:
        :return: callback, if any
        :rtype: callable or None
        """
        self.callbacks.pop(consumer_tag, None)
        return self.cancel_callbacks.pop(consumer_tag, None)

    @synchronized_connection()
    def basic_consume(self, queue='', consumer_tag='', no_local=False, no_ack=False,
                      exclusive=False, nowait=False, callback=None, arguments=None, on_cancel=None):
        """Start a queue consumer

        This method asks the server to start a "consumer", which is a transient request for messages
        from a specific queue. Consumers last as long as the channel they were created on, or until
        the client cancels them.

        * The `consumer_tag` is local to a connection, so two clients can use the same consumer
          tags. But on the same connection, the `consumer_tag` must be unique, or the server must
          raise a 530 NOT ALLOWED connection exception.
        * If `no_ack` is set, the server automatically acknowledges each message on behalf of the
          client.
        * If `exclusive` is set, the client asks for this consumer to have exclusive access to
          the queue. If the server cannot grant exclusive access to the queue because there are
          other consumers active, it must raise a 403 ACCESS REFUSED channel exception.
        * `callback` must be a `Callable(message)` which is called for each messaged delivered by
          the broker. If no callback is specified, messages are quietly discarded; `no_ack` should
          probably be set to True in that case.

        :param str queue: name of queue; if None, refers to last declared queue for this channel
        :param str consumer_tag: consumer tag, local to the connection
        :param bool no_local: if True: do not deliver own messages
        :param bool no_ack: server will not expect an ack for each message
        :param bool exclusive: request exclusive access
        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        :param Callable callback: a callback callable(message) for each delivered message
        :param dict arguments: AMQP method arguments
        :param Callable on_cancel: a callback callable
        :return: consumer tag
        :rtype: str
        """
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(queue)
        args.write_shortstr(consumer_tag)
        args.write_bit(no_local)
        args.write_bit(no_ack)
        args.write_bit(exclusive)
        args.write_bit(nowait)
        args.write_table(arguments or {})
        self._send_method(Method(spec.Basic.Consume, args))

        if not nowait:
            consumer_tag = self.wait(spec.Basic.ConsumeOk)

        self.callbacks[consumer_tag] = callback

        if on_cancel:
            self.cancel_callbacks[consumer_tag] = on_cancel

        if no_ack:
            self.no_ack_consumers.add(consumer_tag)

        return consumer_tag

    def _cb_basic_consume_ok(self, method):
        """Confirm a new consumer

        The server provides the client with a consumer tag, which is used by the client for methods
        called on the consumer at a later stage.

        PARAMETERS:

            consumer_tag: shortstr

                Holds the consumer tag specified by the client or provided by the server.
        """
        args = method.args
        return args.read_shortstr()

    def _cb_basic_deliver(self, method):
        """Notify the client of a consumer message

        This method delivers a message to the client, via a consumer. In the asynchronous message
        delivery model, the client starts a consumer using the Consume method, then the server
        responds with Deliver methods as and when messages arrive for that consumer.

        This method can be called in a "classmethod" style static-context and is done so by
        :meth:`~amqpy.connection.Connection.drain_events()`.

        RULE:

            The server SHOULD track the number of times a message has been delivered to clients and
            when a message is redelivered a certain number of times - e.g. 5 times - without being
            acknowledged, the server SHOULD consider the message to be unprocessable (possibly
            causing client applications to abort), and move the message to a dead letter queue.

        PARAMETERS:

            consumer_tag: shortstr

                consumer tag

                Identifier for the consumer, valid within the current connection.

                RULE:

                    The consumer tag is valid only within the channel from which the consumer was
                    created. I.e. a client
                    MUST NOT create a consumer in one channel and then use it in another.

            delivery_tag: longlong

                server-assigned delivery tag

                The server-assigned and channel-specific delivery tag

                RULE:

                    The delivery tag is valid only within the channel from which the message was
                    received I.e. a
                    client MUST NOT receive a message on one channel and then acknowledge it on
                    another.

                RULE:

                    The server MUST NOT use a zero value for delivery tags Zero is reserved for
                    client use, meaning
                    "all messages so far received".

            redelivered: boolean

                message is being redelivered

                This indicates that the message has been previously delivered to this or another
                client.

            exchange: shortstr

                Specifies the name of the exchange that the message was originally published to.

            routing_key: shortstr

                Message routing key

                Specifies the routing key name specified when the message was published.
        """
        args = method.args
        msg = method.content

        consumer_tag = args.read_shortstr()
        delivery_tag = args.read_longlong()
        redelivered = args.read_bit()
        exchange = args.read_shortstr()
        routing_key = args.read_shortstr()

        msg.channel = self
        msg.delivery_info = {
            'consumer_tag': consumer_tag,
            'delivery_tag': delivery_tag,
            'redelivered': redelivered,
            'exchange': exchange,
            'routing_key': routing_key,
        }

        callback = self.callbacks.get(consumer_tag)
        if callback:
            callback(msg)
        else:
            raise Exception('No callback available for consumer tag: {}'.format(consumer_tag))

    @synchronized_connection()
    def basic_get(self, queue='', no_ack=False):
        """Directly get a message from the `queue`

        This method is non-blocking. If no messages are available on the queue, `None` is returned.

        :param str queue: queue name; leave blank to refer to last declared queue for the channel
        :param bool no_ack: if enabled, the server automatically acknowledges the message
        :return: message, or None if no messages are available on the queue
        :rtype: amqpy.message.Message or None
        """
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(queue)
        args.write_bit(no_ack)
        self._send_method(Method(spec.Basic.Get, args))
        return self.wait_any([spec.Basic.GetOk, spec.Basic.GetEmpty])

    def _cb_basic_get_empty(self, method):
        """Indicate no messages available

        This method tells the client that the queue has no messages
        available for the client.
        """
        args = method.args
        args.read_shortstr()

    def _cb_basic_get_ok(self, method):
        """Provide client with a message

        This method delivers a message to the client following a get method. A message delivered
        by 'get-ok' must be acknowledged unless the no-ack option was set in the get method.

        PARAMETERS:

            delivery_tag: longlong

                server-assigned delivery tag

                The server-assigned and channel-specific delivery tag

                RULE:

                    The delivery tag is valid only within the channel from which the message was
                    received I.e. a
                    client MUST NOT receive a message on one channel and then acknowledge it on
                    another.

                RULE:

                    The server MUST NOT use a zero value for delivery tags Zero is reserved for
                    client use, meaning
                    "all messages so far received".

            redelivered: boolean

                message is being redelivered

                This indicates that the message has been previously delivered to this or another
                client.

            exchange: shortstr

                Specifies the name of the exchange that the message was originally published to.
                If empty, the message
                was published to the default exchange.

            routing_key: shortstr

                Message routing key

                Specifies the routing key name specified when the message was published.

            message_count: long

                number of messages pending

                This field reports the number of messages pending on the queue, excluding the
                message being delivered.
                Note that this figure is indicative, not reliable, and can change arbitrarily as
                messages are added to
                the queue and removed by other clients.
        """
        args = method.args
        msg = method.content

        delivery_tag = args.read_longlong()
        redelivered = args.read_bit()
        exchange = args.read_shortstr()
        routing_key = args.read_shortstr()
        message_count = args.read_long()

        msg.channel = self
        msg.delivery_info = {
            'delivery_tag': delivery_tag,
            'redelivered': redelivered,
            'exchange': exchange,
            'routing_key': routing_key,
            'message_count': message_count
        }
        return msg

    def _basic_publish(self, msg, exchange='', routing_key='', mandatory=False, immediate=False):
        args = AMQPWriter()
        args.write_short(0)
        args.write_shortstr(exchange)
        args.write_shortstr(routing_key)
        args.write_bit(mandatory)
        args.write_bit(immediate)

        self._send_method(Method(spec.Basic.Publish, args, msg))

    @synchronized_connection()
    def basic_publish(self, msg, exchange='', routing_key='', mandatory=False, immediate=False):
        """Publish a message

        This method publishes a message to a specific exchange. The message will be routed to
        queues as defined by the exchange configuration and distributed to any active consumers when
        the transaction, if any, is committed.

        If publisher confirms are enabled, this method will automatically wait to receive an "ack"
        from the server.

        .. note::

            Returned messages are sent back from the server and loaded into
            the `returned_messages` queue of the channel that sent them. In
            order to receive all returned messages, call `loop(0)` on the
            connection object before checking the channel's
            `returned_messages` queue.

        :param msg: message
        :param str exchange: exchange name, empty string means default exchange
        :param str routing_key: routing key
        :param bool mandatory: True: deliver to at least one queue, or return it; False: drop the
            unroutable message
        :param bool immediate: request immediate delivery
        :type msg: amqpy.Message
        """
        self._basic_publish(msg, exchange, routing_key, mandatory, immediate)
        if self.mode == self.CH_MODE_CONFIRM:
            self.wait(spec.Basic.Ack)

    @synchronized_connection()
    def basic_qos(self, prefetch_size=0, prefetch_count=0, a_global=False):
        """Specify quality of service

        This method requests a specific quality of service. The QoS can be specified for the
        current channel or for all channels on the connection. The particular properties and
        semantics of a qos method always depend on the content class semantics. Though the qos
        method could in principle apply to both peers, it is currently meaningful only for the
        server.

        * The client can request that messages be sent in advance so that when the client finishes
          processing a message, the following message is already held locally, rather than needing
          to be sent down the channel. Prefetching gives a performance improvement. This field
          specifies the prefetch window size in octets. The server will send a message in advance
          if it is equal to or smaller in size than the available prefetch size (and also falls
          into other prefetch limits). May be set to zero, meaning "no specific limit", although
          other prefetch limits may still apply. The prefetch-size is ignored if the no-ack option
          is set.
        * The server must ignore `prefetch_size` setting when the client is not processing any
          messages - i.e. the prefetch size does not limit the transfer of single messages to a
          client, only the sending in advance of more messages while the client still has one or
          more unacknowledged messages.
        * The `prefetch_count` specifies a prefetch window in terms of whole messages. This field
          may be used in combination with the prefetch-size field; a message will only be sent in
          advance if both prefetch windows (and those at the channel and connection level) allow
          it. The prefetch-count is ignored if the no-ack option is set.
        * The server may send less data in advance than allowed by the client's specified
          prefetch windows but it must not send more.

        :param int prefetch_size: prefetch window in octets
        :param int prefetch_count: prefetch window in messages
        :param bool a_global: apply to entire connection (default is for current channel only)
        """
        args = AMQPWriter()
        args.write_long(prefetch_size)
        args.write_short(prefetch_count)
        args.write_bit(a_global)
        self._send_method(Method(spec.Basic.Qos, args))
        return self.wait(spec.Basic.QosOk)

    def _cb_basic_qos_ok(self, method):
        """Confirm the requested qos

        This method tells the client that the requested QoS levels could be handled by the server.
        The requested QoS applies to all active consumers until a new QoS is defined.
        """
        pass

    @synchronized_connection()
    def basic_recover(self, requeue=False):
        """Redeliver unacknowledged messages

        This method asks the broker to redeliver all unacknowledged messages on a specified
        channel. Zero or more messages may be redelivered. This method is only allowed on
        non-transacted channels.

        * The server MUST set the redelivered flag on all messages that are resent.
        * The server MUST raise a channel exception if this is called on a transacted channel.

        :param bool requeue: if set, the server will attempt to requeue the message, potentially
            then delivering it to a different subscriber
        """
        args = AMQPWriter()
        args.write_bit(requeue)
        self._send_method(Method(spec.Basic.Recover, args))

    @synchronized_connection()
    def basic_recover_async(self, requeue=False):
        """Redeliver unacknowledged messages (async)

        This method asks the broker to redeliver all unacknowledged messages on a specified
        channel. Zero or more messages may be redelivered. This method is only allowed on
        non-transacted channels.

        * The server MUST set the redelivered flag on all messages that are resent.
        * The server MUST raise a channel exception if this is called on a transacted channel.

        :param bool requeue: if set, the server will attempt to requeue the message, potentially
            then delivering it to a different subscriber
        """
        args = AMQPWriter()
        args.write_bit(requeue)
        self._send_method(Method(spec.Basic.RecoverAsync, args))

    def _cb_basic_recover_ok(self, method):
        """In 0-9-1 the deprecated recover solicits a response
        """
        pass

    @synchronized_connection()
    def basic_reject(self, delivery_tag, requeue):
        """Reject an incoming message

        This method allows a client to reject a message. It can be used to interrupt and cancel
        large incoming messages,
        or return untreatable messages to their original queue.

        * The server SHOULD be capable of accepting and process the Reject method while sending
          message content with a Deliver or Get-Ok method I.e. the server should read and process
          incoming methods while sending output frames. To cancel a partially-send content, the
          server sends a content body frame of size 1 (i.e. with no data except the frame-end
          octet).
        * The server SHOULD interpret this method as meaning that the client is unable to process
          the message at this time.
        * A client MUST NOT use this method as a means of selecting messages to process A
          rejected message MAY be discarded or dead-lettered, not necessarily passed to another
          client.
        * The server MUST NOT deliver the message to the same client within the context of the
          current channel. The recommended strategy is to attempt to deliver the message to an
          alternative consumer, and if that is not possible, to move the message to a dead-letter
          queue. The server MAY use more sophisticated tracking to hold the message on the queue and
          redeliver it to the same client at a later stage.

        :param int delivery_tag: server-assigned channel-specific delivery tag
        :param bool requeue: True: requeue the message; False: discard the message
        """
        args = AMQPWriter()
        args.write_longlong(delivery_tag)
        args.write_bit(requeue)
        self._send_method(Method(spec.Basic.Reject, args))

    def _cb_basic_return(self, method):
        """Return a failed message

        This method returns an undeliverable message that was published with the `immediate` flag
        set, or an unroutable message published with the `mandatory` flag set. The reply code and
        text provide information about the reason that the message was undeliverable.
        """
        args = method.args
        msg = method.content
        self.returned_messages.put(basic_return_t(
            args.read_short(),
            args.read_shortstr(),
            args.read_shortstr(),
            args.read_shortstr(),
            msg,
        ))

    @synchronized_connection()
    def tx_commit(self):
        """Commit the current transaction

        This method commits all messages published and acknowledged in the current transaction. A
        new transaction starts immediately after a commit.
        """
        self._send_method(Method(spec.Tx.Commit))
        return self.wait(spec.Tx.CommitOk)

    def _cb_tx_commit_ok(self, method):
        """Confirm a successful commit

        This method confirms to the client that the commit succeeded. Note that if a commit fails,
        the server raises a channel exception.
        """
        pass

    @synchronized_connection()
    def tx_rollback(self):
        """Abandon the current transaction

        This method abandons all messages published and acknowledged in the current transaction. A
        new transaction starts immediately after a rollback.
        """
        self._send_method(Method(spec.Tx.Rollback))
        return self.wait(spec.Tx.RollbackOk)

    def _cb_tx_rollback_ok(self, method):
        """Confirm a successful rollback

        This method confirms to the client that the rollback succeeded. Note that if an rollback
        fails, the server raises a channel exception.
        """
        pass

    @synchronized_connection()
    def tx_select(self):
        """Select standard transaction mode

        This method sets the channel to use standard transactions. The client must use this method
        at least once on a channel before using the Commit or Rollback methods.

        The channel must not be in publish acknowledge mode. If it is, the server raises a
        :exc:`PreconditionFailed` exception and closes the channel. Note that amqpy will
        automatically reopen the channel, at which point this method can be called again
        successfully.

        :raise PreconditionFailed: if the channel is in publish acknowledge mode
        """
        self._send_method(Method(spec.Tx.Select))
        #self.wait(spec.Tx.SelectOk)
        self.wait(spec.Tx.SelectOk)
        self.mode = self.CH_MODE_TX

    def _cb_tx_select_ok(self, method):
        """Confirm transaction mode

        This method confirms to the client that the channel was successfully set to use standard
        transactions.
        """
        pass

    @synchronized_connection()
    def confirm_select(self, nowait=False):
        """Enable publisher confirms for this channel (RabbitMQ extension)

        The channel must not be in transactional mode. If it is, the server raises a
        :exc:`PreconditionFailed` exception and closes the channel. Note that amqpy will
        automatically reopen the channel, at which point this method can be called again
        successfully.

        :param bool nowait: if set, the server will not respond to the method and the client
            should not wait for a reply
        :raise PreconditionFailed: if the channel is in transactional mode
        """
        args = AMQPWriter()
        args.write_bit(nowait)

        self._send_method(Method(spec.Confirm.Select, args))
        if not nowait:
            self.wait(spec.Confirm.SelectOk)
        self.mode = self.CH_MODE_CONFIRM

    def _cb_confirm_select_ok(self, method):
        """With this method, the broker confirms to the client that the channel is now using
        publisher confirms
        """
        pass

    def _cb_basic_ack_recv(self, method):
        """Callback for receiving a `spec.Basic.Ack`

        This will be called when the server acknowledges a published message (RabbitMQ extension).
        """
        # args = method.args
        # delivery_tag = args.read_longlong()
        # multiple = args.read_bit()

    METHOD_MAP = {
        spec.Channel.OpenOk: _cb_open_ok,
        spec.Channel.Flow: _cb_flow,
        spec.Channel.FlowOk: _cb_flow_ok,
        spec.Channel.Close: _cb_close,
        spec.Channel.CloseOk: _cb_close_ok,
        spec.Exchange.DeclareOk: _cb_exchange_declare_ok,
        spec.Exchange.DeleteOk: _cb_exchange_delete_ok,
        spec.Exchange.BindOk: _cb_exchange_bind_ok,
        spec.Exchange.UnbindOk: _cb_exchange_unbind_ok,
        spec.Queue.DeclareOk: _cb_queue_declare_ok,
        spec.Queue.BindOk: _cb_queue_bind_ok,
        spec.Queue.PurgeOk: _cb_queue_purge_ok,
        spec.Queue.DeleteOk: _cb_queue_delete_ok,
        spec.Queue.UnbindOk: _cb_queue_unbind_ok,
        spec.Basic.QosOk: _cb_basic_qos_ok,
        spec.Basic.ConsumeOk: _cb_basic_consume_ok,
        spec.Basic.Cancel: _cb_basic_cancel_notify,
        spec.Basic.CancelOk: _cb_basic_cancel_ok,
        spec.Basic.Return: _cb_basic_return,
        spec.Basic.Deliver: _cb_basic_deliver,
        spec.Basic.GetOk: _cb_basic_get_ok,
        spec.Basic.GetEmpty: _cb_basic_get_empty,
        spec.Basic.Ack: _cb_basic_ack_recv,
        spec.Basic.RecoverOk: _cb_basic_recover_ok,
        spec.Confirm.SelectOk: _cb_confirm_select_ok,
        spec.Tx.SelectOk: _cb_tx_select_ok,
        spec.Tx.CommitOk: _cb_tx_commit_ok,
        spec.Tx.RollbackOk: _cb_tx_rollback_ok,
    }



from __future__ import absolute_import, division, print_function

__metaclass__ = type
from collections import namedtuple

queue_declare_ok_t = namedtuple('queue_declare_ok_t', ['queue', 'message_count', 'consumer_count'])

basic_return_t = namedtuple('basic_return_t',
                            ['reply_code', 'reply_text', 'exchange', 'routing_key', 'message'])

method_t = namedtuple('method_t', ['class_id', 'method_id'])

#: The default, minimum frame size that both the client and server must be able to handle
FRAME_MIN_SIZE = 4096


class FrameType:
    """This class contains frame-related constants

    METHOD, HEADER, BODY, and HEARTBEAT are all frame type constants which make up the first byte
    of every frame. The END constant is the termination value which is the last byte of every frame.
    """
    METHOD = 1  # method frame
    HEADER = 2  # content header frame
    BODY = 3  # content body frame
    HEARTBEAT = 8  # heartbeat frame
    END = 206  # not actually a frame type; this is the frame terminator byte


class Connection:
    CLASS_ID = 10

    Start = method_t(10, 10)
    StartOk = method_t(10, 11)
    Secure = method_t(10, 20)
    SecureOk = method_t(10, 21)
    Tune = method_t(10, 30)
    TuneOk = method_t(10, 31)
    Open = method_t(10, 40)
    OpenOk = method_t(10, 41)
    Close = method_t(10, 50)
    CloseOk = method_t(10, 51)
    Blocked = method_t(10, 60)
    Unblocked = method_t(10, 61)


class Channel:
    CLASS_ID = 20

    Open = method_t(20, 10)
    OpenOk = method_t(20, 11)
    Flow = method_t(20, 20)
    FlowOk = method_t(20, 21)
    Close = method_t(20, 40)
    CloseOk = method_t(20, 41)


class Exchange:
    CLASS_ID = 40

    Declare = method_t(40, 10)
    DeclareOk = method_t(40, 11)
    Delete = method_t(40, 20)
    DeleteOk = method_t(40, 21)
    Bind = method_t(40, 30)
    BindOk = method_t(40, 31)
    Unbind = method_t(40, 40)
    UnbindOk = method_t(40, 51)


class Queue:
    CLASS_ID = 50

    Declare = method_t(50, 10)
    DeclareOk = method_t(50, 11)
    Bind = method_t(50, 20)
    BindOk = method_t(50, 21)
    Purge = method_t(50, 30)
    PurgeOk = method_t(50, 31)
    Delete = method_t(50, 40)
    DeleteOk = method_t(50, 41)
    Unbind = method_t(50, 50)
    UnbindOk = method_t(50, 51)


class Basic:
    CLASS_ID = 60

    Qos = method_t(60, 10)
    QosOk = method_t(60, 11)
    Consume = method_t(60, 20)
    ConsumeOk = method_t(60, 21)
    Cancel = method_t(60, 30)
    CancelOk = method_t(60, 31)
    Publish = method_t(60, 40)
    Return = method_t(60, 50)
    Deliver = method_t(60, 60)
    Get = method_t(60, 70)
    GetOk = method_t(60, 71)
    GetEmpty = method_t(60, 72)
    Ack = method_t(60, 80)
    Reject = method_t(60, 90)
    RecoverAsync = method_t(60, 100)
    Recover = method_t(60, 110)
    RecoverOk = method_t(60, 111)


class Confirm:
    CLASS_ID = 85

    Select = method_t(85, 10)
    SelectOk = method_t(85, 11)


class Tx:
    CLASS_ID = 90

    Select = method_t(90, 10)
    SelectOk = method_t(90, 11)
    Commit = method_t(90, 20)
    CommitOk = method_t(90, 21)
    Rollback = method_t(90, 30)
    RollbackOk = method_t(90, 31)



import os

from setuptools import setup, find_packages

VERSION = (0, 13, 1)
__version__ = '.'.join(map(str, VERSION[0:3])) + ''.join(VERSION[3:])
__author__ = 'veegee'
__maintainer__ = 'veegee'
__contact__ = 'veegee@veegee.org'
__homepage__ = 'http://github.com/veegee/amqpy'
__docformat__ = 'restructuredtext'

name = 'amqpy'
description = 'an AMQP 0.9.1 client library for Python 2.7 & Python >= 3.2.0'
keywords = ['amqp', 'rabbitmq', 'qpid']

classifiers = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python',
    'Programming Language :: Python :: Implementation :: CPython',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'License :: OSI Approved :: MIT License',
    'Intended Audience :: Developers',
    'Topic :: Internet',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: System :: Networking'
]

package_data = {
    '': ['*.rst', '*.ini', 'AUTHORS', 'LICENSE'],
}


def long_description():
    if os.path.exists('README.rst'):
        with open('README.rst') as f:
            return f.read()
    else:
        return description


setup(
    name=name,
    description=description,
    long_description=long_description(),
    version=__version__,
    author=__author__,
    author_email=__contact__,
    maintainer=__maintainer__,
    url=__homepage__,
    platforms=['any'],
    license='LGPL',
    packages=find_packages(exclude=['ez_setup', 'tests', 'tests.*']),
    package_data=package_data,
    setup_requires=['six>=1.0'],
    install_requires=['six>=1.0'],
    tests_require=['pytest>=2.6'],
    classifiers=classifiers,
    keywords=keywords
)



"""AMQP Connections
"""
from __future__ import absolute_import, division, print_function

__metaclass__ = type
import logging
import socket
from array import array
import pprint
import six
from threading import Event, Thread, Lock

from . import __version__, compat
from .proto import Method
from .method_io import MethodReader, MethodWriter
from .serialization import AMQPWriter
from .abstract_channel import AbstractChannel
from .channel import Channel
from .exceptions import ResourceError, AMQPConnectionError, Timeout, error_for_code
from .transport import create_transport
from . import spec
from .spec import method_t
from .concurrency import synchronized_connection
from .login import login_responses

__all__ = ['Connection']

# client property info that gets sent to the server on connection startup
LIBRARY_PROPERTIES = {
    'product': 'amqpy',
    'product_version': __version__,
    'version': __version__,
    'capabilities': {},
}

log = logging.getLogger('amqpy')

compat.patch()


class Connection(AbstractChannel):
    """The connection class provides methods for a client to establish a network connection to a
    server, and for both peers to operate the connection thereafter
    """

    def __init__(self, host='localhost', port=5672, ssl=None, connect_timeout=None,
                 userid='guest', password='guest', login_method='AMQPLAIN', virtual_host='/',
                 locale='en_US',
                 channel_max=65535, frame_max=131072,
                 heartbeat=0,
                 client_properties=None,
                 on_blocked=None, on_unblocked=None):
        """Create a connection to the specified host

        If you are using SSL, make sure the correct port number is specified (usually 5671), as the
        default of 5672 is for non-SSL connections.

        :param str host: host
        :param int port: port
        :param ssl: dict of SSL options passed to :func:`ssl.wrap_socket()`, None to disable SSL
        :param float connect_timeout: connect timeout
        :param str userid: username
        :param str password: password
        :param str login_method: login method (this is server-specific); default is for RabbitMQ
        :param str virtual_host: virtual host
        :param str locale: locale
        :param int channel_max: maximum number of channels
        :param int frame_max: maximum frame payload size in bytes
        :param float heartbeat: heartbeat interval in seconds, 0 disables heartbeat
        :param client_properties: dict of client properties
        :param on_blocked: callback on connection blocked
        :param on_unblocked: callback on connection unblocked
        :type connect_timeout: float or None
        :type client_properties: dict or None
        :type ssl: dict or None
        :type on_blocked: Callable or None
        :type on_unblocked: Callable or None
        """
        log.debug('amqpy {} Connection.__init__()'.format(__version__))
        self.conn_lock = Lock()

        #: Map of `{channel_id: Channel}` for all active channels
        #:
        #: :type: dict[int, Channel]
        self.channels = {}  # dict of {channel_id int: Channel}

        # the connection object itself is treated as channel 0
        super(Connection, self).__init__(self, 0)  # also sets channels[0] = self

        # instance variables
        #: :type: amqpy.transport.Transport
        self.transport = None
        self.method_reader = None
        self.method_writer = None
        self._wait_tune_ok = None

        # properties set in the start method, after a connection is established
        self.version_major = 0
        self.version_minor = 0
        self.server_properties = {}
        self.mechanisms = []
        self.locales = []

        # properties set in the Tune method
        self.channel_max = channel_max
        self.frame_max = frame_max
        if six.PY2:
            self._avail_channel_ids = array(b'H', range(self.channel_max, 0, -1))
        else:
            self._avail_channel_ids = array('H', range(self.channel_max, 0, -1))
        self._heartbeat_final = 0  # final heartbeat interval after negotiation
        self._heartbeat_server = None

        # save connection parameters
        self._host = host
        self._port = port
        self._connect_timeout = connect_timeout
        self._ssl = ssl
        self._userid = userid
        self._password = password
        self._login_method = login_method
        self._virtual_host = virtual_host
        self._locale = locale
        self._heartbeat_client = heartbeat  # original heartbeat interval value proposed by client
        self._client_properties = client_properties

        # callbacks
        self.on_blocked = on_blocked
        self.on_unblocked = on_unblocked

        # heartbeat
        self._close_event = Event()
        self._heartbeat_thread = None

        self.connect()

    def connect(self):
        """Connect using saved connection parameters

        This method does not need to be called explicitly; it is called by
        the constructor during initialization.

        Note: reconnecting invalidates all declarations (channels, queues,
        consumers, delivery tags, etc.).
        """
        # start the connection; this also sends the connection protocol header
        self.connection = self  # AbstractChannel.connection
        self.transport = create_transport(self._host, self._port, self._connect_timeout,
                                          self.frame_max, self._ssl)

        # create global instances of `MethodReader` and `MethodWriter` which can be used by all
        # channels
        self.method_reader = MethodReader(self.transport)
        self.method_writer = MethodWriter(self.transport, self.frame_max)

        # wait for server to send the 'start' method
        self.wait(spec.Connection.Start)

        # create 'login response' to send to server
        login_response = login_responses[self._login_method](self._userid, self._password)

        # reply with 'start-ok' and connection parameters
        # noinspection PyArgumentList
        client_props = dict(LIBRARY_PROPERTIES, **self._client_properties or {})
        self._send_start_ok(client_props, self._login_method, login_response, self._locale)

        self._wait_tune_ok = True
        while self._wait_tune_ok:
            self.wait_any([spec.Connection.Secure, spec.Connection.Tune])

        self._send_open(self._virtual_host)

        # set up automatic heartbeats, if requested for:
        if self._heartbeat_final:
            self._close_event.clear()
            log.debug('Start automatic heartbeat thread')
            thr = Thread(target=self._heartbeat_run,
                         name='amqp-HeartBeatThread-%s' % id(self))
            thr.daemon = True
            thr.start()
            self._heartbeat_thread = thr

    @property
    def last_heartbeat_recv(self):
        return self.transport.last_heartbeat_received

    @property
    def last_heartbeat_sent(self):
        return self.transport.last_heartbeat_sent

    @property
    def connected(self):
        """Check if connection is connected

        :return: True if connected, else False
        :rtype: bool
        """
        return bool(self.transport and self.transport.connected)

    @property
    def sock(self):
        """Access underlying TCP socket

        :return: socket
        :rtype: socket.socket
        """
        if self.transport and self.transport.sock:
            return self.transport.sock

    @property
    def server_capabilities(self):
        """Get server capabilities

        These properties are set only after successfully connecting.

        :return: server capabilities
        :rtype: dict
        """
        return self.server_properties.get('capabilities') or {}

    @synchronized_connection()
    def channel(self, channel_id=None):
        """Create a new channel, or fetch the channel associated with `channel_id` if specified

        :param channel_id: channel ID number
        :type channel_id: int or None
        :return: Channel
        :rtype: amqpy.channel.Channel
        """
        try:
            return self.channels[channel_id]
        except KeyError:
            return Channel(self, channel_id)

    def send_heartbeat(self):
        """Send a heartbeat to the server
        """
        self.transport.send_heartbeat()

    def is_alive(self):
        """Check if connection is alive

        This method is the primary way to check if the connection is alive.

        Side effects: This method may send a heartbeat as a last resort to check if the connection
        is alive.

        :return: True if connection is alive, else False
        :rtype: bool
        """
        if self.transport:
            return self.transport.is_alive()
        else:
            return False

    def _wait_any(self, timeout=None):
        """Wait for any event on the connection (for any channel)

        When a method is received on the channel, it is delivered to the
        appropriate channel incoming method queue

        :param float timeout: timeout
        :return: method
        :rtype: amqpy.proto.Method
        """
        # check the method queue of each channel
        for ch_id, channel in self.channels.items():
            if channel.incoming_methods:
                return channel.incoming_methods.pop(0)

        # do a blocking read for any incoming method
        method = self.method_reader.read_method(timeout)
        return method

    def drain_events(self, timeout=None):
        """Wait for an event on all channels

        This method should be called after creating consumers in order to
        receive delivered messages and execute consumer callbacks.

        :param timeout: maximum allowed time to wait for an event
        :type timeout: float or None
        :raise amqpy.exceptions.Timeout: if the operation times out
        """
        method = self._wait_any(timeout)

        assert isinstance(method, Method)
        #: :type: amqpy.Channel
        channel = self.channels[method.channel_id]
        return channel.handle_method(method)

    def loop(self, timeout=None):
        """Call :meth:`drain_events` continuously

        - Does not raise Timeout exceptions if a timeout occurs

        :param timeout: maximum allowed time to wait for an event
        :type timeout: float or None
        """
        while True:
            try:
                self.drain_events(timeout)
            except Timeout:
                break

    def close(self, reply_code=0, reply_text='', method_type=method_t(0, 0)):
        """Close connection to the server

        This method performs a connection close handshake with the server, then closes the
        underlying connection.

        If this connection close is due to a client error, the client may provide a `reply_code`,
        `reply_text`, and `method_type` to indicate to the server the reason for closing the
        connection.

        :param int reply_code: the reply code
        :param str reply_text: localized reply text
        :param method_type: if close is triggered by a failing method, this is the method that
            caused it
        :type method_type: amqpy.spec.method_t
        """
        if not self.is_alive():
            # already closed
            log.debug('Already closed')
            return

        # signal to the heartbeat thread to stop sending heartbeats
        if self._heartbeat_final:
            self._close_event.set()
            self._heartbeat_thread.join()
            self._heartbeat_thread = None

        args = AMQPWriter()
        args.write_short(reply_code)
        args.write_shortstr(reply_text)
        args.write_short(method_type.class_id)
        args.write_short(method_type.method_id)
        self._send_method(Method(spec.Connection.Close, args))
        return self.wait_any([spec.Connection.Close, spec.Connection.CloseOk])

    def _heartbeat_run(self):
        # `is_alive()` sends heartbeats if the connection is alive
        while self.is_alive():
            # `close` is set to true if the `close_event` is signalled
            close = self._close_event.wait(self._heartbeat_final / 1.5)
            if close:
                break

    def _close(self):
        try:
            self.transport.close()

            channels = [x for x in self.channels.values() if x is not self]
            for ch in channels:
                # noinspection PyProtectedMember
                ch._close()
        except socket.error:
            pass  # connection already closed on the other end
        finally:
            self.transport = self.connection = None
            self.channels = {0: self}  # reset the channels state

    def _get_free_channel_id(self):
        """Get next free channel ID

        :return: next free channel_id
        :rtype: int
        """
        try:
            return self._avail_channel_ids.pop()
        except IndexError:
            raise ResourceError('No free channel ids, current={0}, channel_max={1}'.format(
                len(self.channels), self.channel_max), spec.Channel.Open)

    def _claim_channel_id(self, channel_id):
        """Claim channel ID

        :param channel_id: channel ID
        :type channel_id: int
        """
        try:
            return self._avail_channel_ids.remove(channel_id)
        except ValueError:
            raise AMQPConnectionError('Channel {} already open'.format(channel_id))

    def _cb_close(self, method):
        """Handle received connection close

        This method indicates that the sender (server) wants to close the connection. This may be
        due to internal conditions (e.g. a forced shut-down) or due to an error handling a specific
        method, i.e. an exception. When a close is due to an exception, the sender provides the
        class and method id of the method which caused the exception.
        """
        args = method.args
        reply_code = args.read_short()  # the AMQP reply code
        reply_text = args.read_shortstr()  # the localized reply text
        class_id = args.read_short()  # class_id of method
        method_id = args.read_short()  # method_id of method

        self._send_close_ok()  # send a close-ok to the server, to confirm that we've
        # acknowledged the close request

        method_type = method_t(class_id, method_id)
        raise error_for_code(reply_code, reply_text, method_type, AMQPConnectionError,
                             self.channel_id)

    def _cb_blocked(self, method):
        """RabbitMQ Extension
        """
        reason = method.args.read_shortstr()
        if callable(self.on_blocked):
            # noinspection PyCallingNonCallable
            return self.on_blocked(reason)

    def _cb_unblocked(self, method):
        assert method
        if callable(self.on_unblocked):
            # noinspection PyCallingNonCallable
            return self.on_unblocked()

    def _send_close_ok(self):
        """Confirm a connection close that has been requested by the server

        This method confirms a Connection.Close method and tells the recipient that it is safe to
        release resources for the connection and close the socket. RULE: A peer that detects a
        socket closure without having received a Close-Ok handshake method SHOULD log the error.
        """
        self._send_method(Method(spec.Connection.CloseOk))
        self._close()

    def _cb_close_ok(self, method):
        """Confirm a connection close

        This method is called when the server send a close-ok in response to our close request. It
        is now safe to close the underlying connection.
        """
        assert method
        self._close()

    def _send_open(self, virtual_host, capabilities=''):
        """Open connection to virtual host

        This method opens a connection to a virtual host, which is a collection of resources, and
        acts to separate multiple application domains within a server. RULE: The client MUST open
        the context before doing any work on the connection.

        :param virtual_host: virtual host path
        :param capabilities: required capabilities
        :type virtual_host: str
        :type capabilities: str
        """
        args = AMQPWriter()
        args.write_shortstr(virtual_host)
        args.write_shortstr(capabilities)
        args.write_bit(False)
        self._send_method(Method(spec.Connection.Open, args))
        return self.wait(spec.Connection.OpenOk)

    def _cb_open_ok(self, method):
        """Signal that the connection is ready

        This method signals to the client that the connection is ready for use.
        """
        assert method
        log.debug('Open OK')

    def _cb_secure(self, method):
        """Security mechanism challenge

        The SASL protocol works by exchanging challenges and responses until both peers have
        received sufficient information to authenticate each other This method challenges the
        client to provide more information.

        PARAMETERS:
            challenge: longstr
                security challenge data
                Challenge information, a block of opaque binary data passed to the security
                mechanism.
        """
        challenge = method.args.read_longstr()
        assert challenge

    def _send_secure_ok(self, response):
        """Security mechanism response

        This method attempts to authenticate, passing a block of SASL data for the security
        mechanism at the server side.

        PARAMETERS:
            response: longstr
                security response data
                A block of opaque data passed to the security mechanism. The contents of this data
                are defined by the SASL security mechanism.
        """
        args = AMQPWriter()
        args.write_longstr(response)
        self._send_method(Method(spec.Connection.SecureOk, args))

    def _cb_start(self, method):
        """Start connection negotiation callback

        This method starts the connection negotiation process by telling the client the protocol
        version that the server proposes, along with a list of security mechanisms which the client
        can use for authentication.

        RULE: If the client cannot handle the protocol version suggested by the server it MUST close
        the socket connection.

        RULE: The server MUST provide a protocol version that is lower than or equal to that
        requested by the client in the protocol header. If the server cannot support the specified
        protocol it MUST NOT send this method, but MUST close the socket connection.

        PARAMETERS:
            version_major: octet
                protocol major version
                The protocol major version that the server agrees to use, which cannot be higher
                than the client's major version.
            version_minor: octet
                protocol major version
                The protocol minor version that the server agrees to use, which cannot be higher
                than the client's minor version.
            server_properties: table
                server properties
            mechanisms: longstr
                available security mechanisms
                A list of the security mechanisms that the server supports, delimited by spaces.
                Currently ASL supports these mechanisms: PLAIN.
            locales: longstr
                available message locales
                A list of the message locales that the server supports, delimited by spaces The
                locale defines the language in which the server will send reply texts.
                RULE:
                    All servers MUST support at least the en_US locale.
        """
        args = method.args
        self.version_major = args.read_octet()
        self.version_minor = args.read_octet()
        self.server_properties = args.read_table()
        self.mechanisms = args.read_longstr().split(' ')
        self.locales = args.read_longstr().split(' ')

        properties = pprint.pformat(self.server_properties)
        log.debug('Start from server')
        log.debug('Version: {}.{}'.format(self.version_major, self.version_minor))
        log.debug('Server properties:\n{}'.format(properties))
        log.debug('Security mechanisms: {}'.format(self.mechanisms))
        log.debug('Locales: {}'.format(self.locales))

    def _send_start_ok(self, client_properties, mechanism, response, locale):
        """Select security mechanism and locale

        This method selects a SASL security mechanism. ASL uses SASL (RFC2222) to negotiate
        authentication and encryption.

        PARAMETERS:
            client_properties: table
                client properties
            mechanism: shortstr
                selected security mechanism
                A single security mechanisms selected by the client, which must be one of those
                specified by the server.
                RULE:
                    The client SHOULD authenticate using the highest- level security profile it
                    can handle from the list
                    provided by the server.
                RULE:
                    The mechanism field MUST contain one of the security mechanisms proposed by
                    the server in the Start
                    method. If it doesn't, the server MUST close the socket.
            response: longstr
                security response data
                A block of opaque data passed to the security mechanism. The contents of this
                data are defined by the
                SASL security mechanism For the PLAIN security mechanism this is defined as a
                field table holding two
                fields, LOGIN and PASSWORD.
            locale: shortstr
                selected message locale
                A single message local selected by the client, which must be one of those
                specified by the server.
        """
        if self.server_capabilities.get('consumer_cancel_notify'):
            if 'capabilities' not in client_properties:
                client_properties['capabilities'] = {}
            client_properties['capabilities']['consumer_cancel_notify'] = True
        if self.server_capabilities.get('connection.blocked'):
            if 'capabilities' not in client_properties:
                client_properties['capabilities'] = {}
            client_properties['capabilities']['connection.blocked'] = True
        args = AMQPWriter()
        args.write_table(client_properties)
        args.write_shortstr(mechanism)
        args.write_longstr(response)
        args.write_shortstr(locale)
        self._send_method(Method(spec.Connection.StartOk, args))

    def _cb_tune(self, method):
        """Handle received "tune" method

        This method is the handler for receiving a "tune" method. `channel_max` and `frame_max`
        are set to the lower
        of the values proposed by each party.

        PARAMETERS:
            channel_max: short
                proposed maximum channels
                The maximum total number of channels that the server allows per connection. Zero
                means that the server
                does not impose a fixed limit, but the number of allowed channels may be limited
                by available server
                resources.
            frame_max: long
                proposed maximum frame size
                The largest frame size that the server proposes for the connection. The client
                can negotiate a lower
                value Zero means that the server does not impose any specific limit but may
                reject very large frames
                if it cannot allocate resources for them.
                RULE:
                    Until the frame-max has been negotiated, both peers MUST accept frames of up
                    to 4096 octets large.
                    The minimum non-zero value for the frame-max field is 4096.
            heartbeat: short
                desired heartbeat delay
                The delay, in seconds, of the connection heartbeat that the server wants Zero
                means the server does
                not want a heartbeat.
        """
        args = method.args
        client_heartbeat = self._heartbeat_client or 0
        # maximum number of channels that the server supports
        self.channel_max = min(args.read_short(), self.channel_max)
        # largest frame size the server proposes for the connection
        self.frame_max = min(args.read_long(), self.frame_max)
        self.method_writer.frame_max = self.frame_max
        # heartbeat interval proposed by server
        self._heartbeat_server = args.read_short() or 0

        # negotiate the heartbeat interval to the smaller of the specified values
        if self._heartbeat_server == 0 or client_heartbeat == 0:
            self._heartbeat_final = max(self._heartbeat_server, client_heartbeat)
        else:
            self._heartbeat_final = min(self._heartbeat_server, client_heartbeat)

        # Ignore server heartbeat if client_heartbeat is disabled
        if not self._heartbeat_client:
            self._heartbeat_final = 0

        self._send_tune_ok(self.channel_max, self.frame_max, self._heartbeat_final)

    def _send_tune_ok(self, channel_max, frame_max, heartbeat):
        """Negotiate connection tuning parameters

        This method sends the client's connection tuning parameters to the server. Certain fields
        are negotiated, others provide capability information.

        PARAMETERS:
            channel_max: short
                negotiated maximum channels
                The maximum total number of channels that the client will use per connection.
                May not be higher than
                the value specified by the server.
                RULE:
                    The server MAY ignore the channel-max value or MAY use it for tuning its
                    resource allocation.
            frame_max: long
                negotiated maximum frame size
                The largest frame size that the client and server will use for the connection.
                Zero means that the
                client does not impose any specific limit but may reject very large frames if it
                cannot allocate
                resources for them. Note that the frame-max limit applies principally to content
                frames, where large
                contents can be broken into frames of arbitrary size.
                RULE:
                    Until the frame-max has been negotiated, both peers must accept frames of up
                    to 4096 octets large.
                    The minimum non-zero value for the frame- max field is 4096.
            heartbeat: short
                desired heartbeat delay
                The delay, in seconds, of the connection heartbeat that the client wants. Zero
                means the client does not
                want a heartbeat.
        """
        args = AMQPWriter()
        args.write_short(channel_max)
        args.write_long(frame_max)
        args.write_short(heartbeat or 0)
        self._send_method(Method(spec.Connection.TuneOk, args))
        self._wait_tune_ok = False

    METHOD_MAP = {
        spec.Connection.Start: _cb_start,
        spec.Connection.Secure: _cb_secure,
        spec.Connection.Tune: _cb_tune,
        spec.Connection.OpenOk: _cb_open_ok,
        spec.Connection.Close: _cb_close,
        spec.Connection.CloseOk: _cb_close_ok,
        spec.Connection.Blocked: _cb_blocked,
        spec.Connection.Unblocked: _cb_unblocked,
    }



"""
This module's purpose is to provide compatibility for Python 3.2 and PyPy3 (which is based on Python
3.2).

The primary method of providing compatibility is by backporting Python 3.3+ features through
monkey-patching. However, this is only done when safe, i.e., when monkey-patching does not change
the behaviour of existing objects.
"""
from __future__ import absolute_import, division, print_function

__metaclass__ = type

import sys
import time
import six

if six.PY2:
    import __builtin__
elif six.PY3:
    import builtins

if sys.version_info < (3, 3):
    # add TimeoutError exception class
    class TimeoutError(OSError):
        pass


def patch():
    if sys.version_info >= (3, 3):
        return

    if not hasattr(time, 'monotonic'):
        from .support import monotonic
        time.monotonic = monotonic.monotonic

    time.perf_counter = time.clock
    if six.PY2:
        __builtin__.TimeoutError = TimeoutError
    elif six.PY3:
        builtins.TimeoutError = TimeoutError
